import collections
import pathlib

import click
import dulwich.objectspec
import dulwich.repo
import phabricator


def get_hash_list(path, old_reference, new_reference):
    repository = dulwich.repo.Repo(path)

    old_hash = dulwich.objectspec.parse_commit(
        repo=repository, committish=old_reference
    )
    new_hash = dulwich.objectspec.parse_commit(
        repo=repository, committish=new_reference
    )

    old_walker = repository.get_walker(include=[old_hash.id])
    old_parents = next(iter(old_walker)).commit.parents

    walker = repository.get_walker(
        include=[new_hash.id],
        exclude=old_parents,
    )
    hashes = [entry.commit.id for entry in walker]

    return hashes


def get_all_commits(phab, hashes):
    after = None
    collected = []

    while True:
        result = phab.request(
            "diffusion.commit.search",
            {
                "after": after,
                "constraints": {
                    "identifiers": [hash.decode("ascii") for hash in hashes],
                },
            },
        )
        collected.extend(result["data"])
        after = result["cursor"]["after"]
        if after is None:
            break

    result = phab.request(
        "phid.lookup",
        {
            "names": [
                *[data["phid"] for data in collected],
                *[data["fields"]["repositoryPHID"] for data in collected],
            ],
        },
    )

    for data in collected:
        data["uri"] = result[data["phid"]]["uri"]
        data["fullName"] = result[data["fields"]["repositoryPHID"]]["fullName"]

    return collected


def create_command(default_phabricator_url=None):
    phabricator_option_extras = {}

    if default_phabricator_url is not None:
        phabricator_option_extras["default"] = default_phabricator_url

    @click.command()
    @click.option(
        "--target",
        "targets",
        help="path to the repository root, old reference, new reference",
        multiple=True,
        type=(
            click.Path(exists=True, file_okay=False, resolve_path=True),
            str,
            str,
        ),
        required=True,
    )
    @click.option(
        "--phabricator-url",
        help="The base URL for the Phabricator server.",
        type=str,
        required=True,
        **phabricator_option_extras,
    )
    @click.option(
        "--api-token",
        help="""
            Your phabricator API token.
            These are managed under your user settings then 'Conduit API Tokens'.
        """,
        type=str,
        required=True,
    )
    def cli(targets, phabricator_url, api_token):
        all_hashes = []

        for target_path, old_reference, new_reference in targets:
            target_path = pathlib.Path(target_path)

            all_hashes.extend(
                get_hash_list(
                    path=target_path,
                    old_reference=old_reference,
                    new_reference=new_reference,
                )
            )

        phab = phabricator.Phabricator(
            phabricator_url,
            "altendky",
            token=api_token,
        )

        results = collections.defaultdict(list)
        commits = get_all_commits(phab=phab, hashes=all_hashes)

        for commit in commits:
            fields = commit["fields"]
            full_hash = fields["identifier"]
            audit_status = fields["auditStatus"]["value"]
            full_name = commit["fullName"]
            uri = commit["uri"]

            link = f"[[ {uri} | {full_name} {full_hash} ]]"

            results[audit_status].append(link)

        sorted_results = dict(sorted(results.items()))
        for status, links in sorted_results.items():
            print(f"- {status}: {len(links)}")

        print()

        for status, links in sorted_results.items():
            print(f"- {status}:")

            for link in links:
                print(f"  - {link}")

            print()

    return cli



import attr
import bisect
import collections
import itertools

import epyqlib.utils.general


class ChunkExistsError(Exception):
    pass


class ChunkNotFoundError(Exception):
    pass


class ByteLengthError(Exception):
    pass


@attr.s(frozen=True)
class Subscriber:
    callback = attr.ib()
    reference = attr.ib()


@attr.s
class Cache:
    _chunks = attr.ib(init=False, default=attr.Factory(list))
    _chunks_set = attr.ib(init=False, default=attr.Factory(set), repr=False)
    _subscribers = attr.ib(init=False, default=attr.Factory(dict))
    _bits_per_byte = attr.ib(default=8)

    def __attrs_post_init__(self):
        self.address_to_chunks = collections.defaultdict(set)

    def __repr__(self):
        return object.__repr__(self)

    def __str__(self):
        return object.__str__(self)

    def add(self, chunk):
        if chunk in self._chunks_set:
            raise ChunkExistsError(chunk)

        bisect.insort_left(self._chunks, chunk)
        self._chunks_set.add(chunk)
        self._subscribers[chunk] = set()

        for address in chunk.addresses():
            self.address_to_chunks[address].add(chunk)

    def subscribe(self, subscriber, chunk, reference=None):
        if chunk not in self._chunks_set:
            raise ChunkNotFoundError(chunk)

        s = Subscriber(callback=subscriber, reference=reference)
        self._subscribers[chunk].add(s)

    # def unsubscribe(self, subscriber, chunk, reference=None):
    #     to_discard = []
    #     for subscriber in self._subscribers[chunk]:
    #         if
    #         .discard(subscriber)

    def unsubscribe_by_reference(self, reference, chunk=None):
        chunks = self._subscribers.items() if chunk is None else [chunk]

        to_discard = set()
        for chunk, subscribers in chunks:
            for subscriber in subscribers:
                # TODO: perhaps should be `is` not `==`?
                if subscriber.reference == reference:
                    to_discard.add(subscriber)

            self._subscribers[chunk] -= to_discard

    def unsubscribe_all(self, *chunks):
        if len(chunks) == 0:
            chunks = self._chunks

        for chunk in chunks:
            self._subscribers[chunk] = set()

    def update(self, update_chunk):

        intersected_chunks = set()
        for address in update_chunk.addresses():
            intersected_chunks |= self.address_to_chunks[address]

        for chunk in intersected_chunks:
            chunk.update(update_chunk)

            for subscriber in self._subscribers[chunk]:
                subscriber.callback(chunk._bytes)

    def chunk_from_variable(self, variable, reference=None):
        data = bytearray([0] * variable.type.bytes * (self._bits_per_byte // 8))
        if reference is None:
            reference = variable
        return Chunk(
            address=variable.address,
            bytes=data,
            bits_per_byte=self._bits_per_byte,
            reference=reference,
        )

    def new_chunk(self, address, bytes, reference=None):
        return Chunk(
            address=address,
            bytes=bytes,
            bits_per_byte=self._bits_per_byte,
            reference=reference,
        )

    def contiguous_chunks(self):
        chunks = []

        addresses = set()
        for chunk in self._chunks:
            addresses.update(
                chunk._address + offset
                for offset in range(len(chunk._bytes) // (self._bits_per_byte // 8))
            )

        gen = epyqlib.utils.general.generate_ranges(sorted(addresses))
        for start, end in gen:
            print(start, end)
            chunks.append(
                self.new_chunk(
                    address=start,
                    bytes=b"\x00" * (end - start + 1) * (self._bits_per_byte // 8),
                )
            )

        return chunks


@attr.s(hash=False)
class Chunk:
    _address = attr.ib()
    _bytes = attr.ib(converter=bytearray)
    _bits_per_byte = attr.ib(default=8)
    reference = attr.ib(cmp=False, default=None)

    def __len__(self):
        # TODO: maybe...
        #       return len(self._bytes) // (self._bits_per_byte // 8)
        return len(self._bytes)

    def addresses(self):
        start, end = self.bounds()
        return range(start, end)

    def bounds(self):
        # TODO: is this returning one too long?
        return (self._address, self._address + len(self) // (self._bits_per_byte // 8))

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented

        return self._address < other._address or (
            self._address == other._address and len(self) < len(other)
        )

    def set_bytes(self, new_bytes):
        if len(new_bytes) != len(self):
            raise ByteLengthError("A chunk's byte array may not change length")

        self._bytes = bytearray(new_bytes)

    def overlaps_address_length(self, address, length):
        # TODO: CAMPid 4317784316796754167954114314396
        my_bounds = self.bounds()
        address_bounds = (address, address + length)

        start = max(my_bounds[0], address_bounds[0])
        end = min(my_bounds[1], address_bounds[1])

        _length = end - start

        return _length >= 1

    def update(self, chunk):
        # TODO: CAMPid 4317784316796754167954114314396
        my_bounds = self.bounds()
        chunk_bounds = chunk.bounds()

        start = max(my_bounds[0], chunk_bounds[0])
        end = min(my_bounds[1], chunk_bounds[1])

        _length = end - start

        if _length >= 1:
            slice_start = (start - self._address) * (self._bits_per_byte // 8)
            slice_end = slice_start + (_length * self._bits_per_byte // 8)
            self_slice = slice(slice_start, slice_end)

            slice_start = (start - chunk._address) * (self._bits_per_byte // 8)
            slice_end = slice_start + (_length * self._bits_per_byte // 8)
            chunk_slice = slice(slice_start, slice_end)

            self._bytes[self_slice] = chunk._bytes[chunk_slice]

            return True

        return False


def testit(filename):
    import epyqlib.cmemoryparser
    import functools
    import random

    names, variables, bits_per_byte = epyqlib.cmemoryparser.process_file(filename)

    cache = Cache(bits_per_byte=bits_per_byte)

    for variable in variables:
        chunk = cache.chunk_from_variable(variable)
        cache.add(chunk)
        partial = functools.partial(testit_updated, variable)
        cache.subscribe(subscriber=partial, chunk=chunk)
        print("added {}: {}".format(variable.name, chunk))

    chunk = cache.chunk_from_variable(names["Vholdoff"])
    new_bytes = [random.randint(0, 255) for _ in range(len(chunk))]
    chunk.set_bytes(new_bytes)
    print("sending update: {}".format(chunk))
    cache.update(chunk)

    testStruct12 = names["testStruct12"]
    TestStruct12 = epyqlib.cmemoryparser.base_type(testStruct12)
    members = ["sA", "m10"]
    member, offset = TestStruct12.member_and_offset(members)

    def all_set_byte():
        return 255

    def random_byte():
        return random.randint(0, 255)

    chunk = cache.new_chunk(
        address=testStruct12.address + offset,
        bytes=[random_byte() for _ in range(member.bytes * (bits_per_byte // 8))],
    )
    print("sending update: {}".format(chunk))
    cache.update(chunk)


def testit_updated(variable, bytes):
    print(
        "{} was updated: {} -> {}".format(variable.name, bytes, variable.unpack(bytes))
    )


if __name__ == "__main__":
    import sys

    sys.exit(testit(sys.argv[1]))



import pathlib
from datetime import datetime
from enum import Enum

from epyqlib.tabs.files.log_manager import PendingLog
from typing import Dict

import PyQt5.uic
import attr
from PyQt5.QtCore import Qt, QPoint
from PyQt5.QtGui import QColor, QBrush, QTextCursor, QFont
from PyQt5.QtWidgets import (
    QPushButton,
    QTreeWidget,
    QTreeWidgetItem,
    QLineEdit,
    QLabel,
    QPlainTextEdit,
    QGridLayout,
    QMenu,
    QTextEdit,
    QAction,
)
from twisted.internet.defer import ensureDeferred, inlineCallbacks

from epyqlib.utils.twisted import errbackhook as open_error_dialog

# noinspection PyUnreachableCode
if False:  # Tell the editor about the type, but don't invoke a cyclic depedency
    from epyqlib.device import DeviceInterface


from epyqlib.tabs.files.graphql import InverterNotFoundException

Ui, UiBase = PyQt5.uic.loadUiType(
    pathlib.Path(__file__).with_suffix(".ui"),
)


class _Sections:
    params: QTreeWidgetItem
    pmvs: QTreeWidgetItem
    firmware: QTreeWidgetItem
    raw_logs: QTreeWidgetItem
    other: QTreeWidgetItem


class Cols:
    filename = 0
    local = 1
    web = 2
    association = 3
    creator = 4
    uploaded_at = 5
    version = 6
    description = 7


class Relationships(Enum):
    inverter = QBrush(QColor(187, 187, 187))  # Gray
    model = QBrush(QColor(112, 173, 71))  # Green
    customer = QBrush(QColor(143, 170, 220))  # Blue
    site = QBrush(QColor(255, 208, 64))  # Yellow


def get_keys(obj):
    return [key for key in obj.__dict__.keys() if not key.startswith("__")]


def get_values(obj):
    return [obj.__dict__[key] for key in get_keys(obj)]


@attr.s
class FilesView(UiBase):
    fa_check = ""
    fa_cross = ""
    fa_wifi = ""
    fa_question = ""

    fontawesome = QFont("fontawesome")
    color_black = QColor("black")
    color_blue = QColor("#1E93F6")
    color_green = QColor("green")
    color_gray = QColor("gray")
    color_red = QColor("red")

    device_interface: "DeviceInterface" = attr.ib(init=False)
    pending_log_rows: Dict[str, QTreeWidgetItem] = {}  # Hash -> Row

    time_format = "%m/%d %I:%M%p "

    ui = attr.ib(factory=Ui)

    @classmethod
    def build(cls):
        instance = cls()
        instance.setup_ui()

        return instance

    @classmethod
    def qt_build(cls, parent):
        instance = cls.build()
        instance.setParent(parent)

        return instance

    def __attrs_post_init__(self):
        super().__init__()

    # noinspection PyAttributeOutsideInit
    def setup_ui(self):
        self.ui.setupUi(self)

        from .files_controller import FilesController

        self.controller = FilesController(self)
        self.controller.setup()

    def set_device_interface(self, device_interface):
        self.controller.device_interface_set(device_interface)

    def tab_selected(self):
        ensureDeferred(self.controller.tab_selected()).addErrback(open_error_dialog)

    def on_bus_status_changed(self):
        ensureDeferred(self.controller.on_bus_status_changed()).addErrback(
            open_error_dialog
        )

    ### Setup methods
    # noinspection PyAttributeOutsideInit
    def bind(self):
        self._log_text = ""
        self._current_file_id: str = None

        self.section_headers = _Sections()

        self.root: QGridLayout = self.ui.gridLayout

        self.lbl_login_status: QLabel = self.ui.lbl_login_status
        self.btn_login: QPushButton = self.ui.login
        self.btn_logout: QPushButton = self.ui.logout

        self.lbl_serial_number: QLabel = self.ui.lbl_serial_number
        self.serial_number: QLineEdit = self.ui.serial_number
        self.serial_number_from_parameters: QPushButton = (
            self.ui.serial_number_from_parameters
        )
        self.inverter_error: QLabel = self.ui.lbl_inverter_error

        self.files_grid: QTreeWidget = self.ui.files_grid

        self.assigned_by: QLineEdit = self.ui.assigned_by
        self.assigned_time: QLineEdit = self.ui.assigned_time
        self.description: QLineEdit = self.ui.description
        self.filename: QLineEdit = self.ui.filename
        self.upload_time: QLineEdit = self.ui.upload_time
        self.version: QLineEdit = self.ui.version

        self.notes: QPlainTextEdit = self.ui.notes
        self.btn_save_notes: QPushButton = self.ui.save_notes
        self.btn_reset_notes: QPushButton = self.ui.reset_notes

        self.lbl_last_sync: QLabel = self.ui.last_sync
        self.btn_sync_now: QPushButton = self.ui.sync_now
        self.btn_sync_all: QPushButton = self.ui.sync_all

        self.event_log: QTextEdit = self.ui.event_log

        # Bind click events
        self.btn_login.clicked.connect(self._login_clicked)
        self.btn_logout.clicked.connect(self._logout_clicked)
        self.serial_number.returnPressed.connect(self._serial_number_entered)
        self.serial_number_from_parameters.clicked.connect(
            self._serial_number_from_parameters_clicked
        )

        self.files_grid.itemClicked.connect(self.controller.file_item_clicked)

        self.btn_sync_now.clicked.connect(self._sync_now_clicked)
        self.btn_sync_all.clicked.connect(self._sync_all_clicked)

        self.notes.textChanged.connect(self._notes_changed)
        self.description.textChanged.connect(self._notes_changed)
        self.btn_save_notes.clicked.connect(self._save_notes_clicked)
        self.btn_reset_notes.clicked.connect(self._reset_notes)

        # Debug button
        self.btn_debug: QPushButton = self.ui.btn_debug
        self.btn_debug.setVisible(False)
        self.btn_debug.clicked.connect(self._debug_clicked)

        # Set initial state

    def populate_tree(self):
        self.files_grid.setAlternatingRowColors(True)

        self.files_grid.setHeaderLabels(get_keys(Cols))
        self.files_grid.setColumnWidth(Cols.filename, 250)
        self.files_grid.setColumnWidth(Cols.local, 35)
        self.files_grid.setColumnWidth(Cols.web, 30)
        self.files_grid.setColumnWidth(Cols.association, 150)
        self.files_grid.setColumnWidth(Cols.description, 500)
        self.files_grid.setColumnWidth(Cols.uploaded_at, 125)

        def make_entry(caption):
            val = QTreeWidgetItem(self.files_grid, [caption])
            val.setExpanded(True)
            return val

        self.section_headers.params = make_entry("Parameter Files")
        self.section_headers.pmvs = make_entry("Value Sets")
        self.section_headers.firmware = make_entry("Firmware")
        self.section_headers.raw_logs = make_entry("Fault Logs")
        self.section_headers.other = make_entry("Other files")

    def initialize_ui(self):
        self.files_grid.setContextMenuPolicy(Qt.CustomContextMenu)
        self.files_grid.customContextMenuRequested.connect(self._render_context_menu)

        self.btn_sync_now.setDisabled(True)

        self.serial_number.setReadOnly(
            False
        )  # TODO: Link to whether or not they have admin access
        self.show_inverter_error(None)

    def disable_serial_number_input(self, disable: bool):
        self.serial_number.setReadOnly(disable)

    def _remove_all_children(self, parent: QTreeWidgetItem):
        while parent.childCount() > 0:
            parent.removeChild(parent.child(0))

    def sort_grid_items(self):
        self.section_headers.raw_logs.sortChildren(
            Cols.uploaded_at, Qt.SortOrder.AscendingOrder
        )
        self.section_headers.params.sortChildren(
            Cols.filename, Qt.SortOrder.AscendingOrder
        )
        self.section_headers.other.sortChildren(
            Cols.filename, Qt.SortOrder.AscendingOrder
        )

    def set_serial_number(self, serial_number: str):
        self.serial_number.setText(serial_number)

    def inverter_error_handler(self, error):
        if (
            error.type is InverterNotFoundException
        ):  # Twisted wraps errors in its own class
            self.show_inverter_error("Error: Inverter ID not found.")
        else:
            raise error

    def get_parent_for_association_type(self, association_type: str):
        parents = {
            "firmware": self.section_headers.firmware,
            "log": self.section_headers.raw_logs,
            "other": self.section_headers.other,
            "parameter": self.section_headers.params,
            "pmvs": self.section_headers.pmvs,
        }

        return parents[association_type.lower()]

    def ensure_correct_parent_for_row(self, row: QTreeWidgetItem, type: str):
        new_parent = self.get_parent_for_association_type(type)
        if row.parent() is not new_parent:
            row.parent().removeChild(row)
            new_parent.addChild(row)

    def attach_row_to_parent(self, association_type: str, filename):
        parent = self.get_parent_for_association_type(association_type)
        row = QTreeWidgetItem(parent, [filename])
        row.setFont(Cols.local, self.fontawesome)
        row.setFont(Cols.web, self.fontawesome)

        row.setTextAlignment(Cols.local, Qt.AlignCenter)
        row.setTextAlignment(Cols.web, Qt.AlignCenter)

        return row

    def remove_row(self, row: QTreeWidgetItem):
        row.parent().removeChild(row)

    def render_association_to_row(self, association, row: QTreeWidgetItem):
        uploaded = association["file"]["createdAt"][:-1]  # Trim trailing "Z"
        uploaded = datetime.fromisoformat(uploaded)

        row.setText(Cols.filename, association["file"]["filename"])
        row.setText(Cols.version, association["file"]["version"])
        row.setText(Cols.uploaded_at, uploaded.strftime(self.time_format))
        row.setText(Cols.description, association["file"]["description"])

        if association["file"]["owner"] == "epc":
            font: QFont = row.font(Cols.creator)
            font.setBold(True)
            row.setFont(Cols.creator, font)
            row.setForeground(Cols.creator, self.color_blue)
            row.setFont(Cols.creator, row.font(Cols.creator))
            row.setText(Cols.creator, "EPC Power")
        else:
            row.setFont(Cols.creator, row.font(Cols.uploaded_at))
            row.setForeground(Cols.creator, self.color_black)
            row.setText(Cols.creator, association["file"].get("createdBy"))

        if association.get("model"):
            model_name = " " + association["model"]["name"]

            if association.get("customer"):
                relationship = Relationships.customer
                rel_text = association["customer"]["name"] + "," + model_name
            elif association.get("site"):
                relationship = Relationships.site
                rel_text = association["site"]["name"] + "," + model_name
            else:
                relationship = Relationships.model
                rel_text = "All" + model_name
        else:
            relationship = Relationships.inverter
            rel_text = "SN: " + association["inverter"]["serialNumber"]

        self.show_relationship(row, relationship, rel_text)

    def add_new_pending_log_row(self, log: PendingLog, ctime: datetime):
        row = self.attach_row_to_parent("log", log.filename)

        self.show_check_icon(row, Cols.local)
        self.show_question_icon(row, Cols.web)

        self.show_relationship(row, Relationships.inverter, f"SN: {log.serial_number}")
        row.setText(Cols.uploaded_at, ctime.strftime(self.time_format))
        row.setText(Cols.creator, log.username)

        self.pending_log_rows[log.hash] = row

    def show_relationship(
        self, row: QTreeWidgetItem, relationship: Relationships, rel_text: str
    ):
        row.setBackground(Cols.association, relationship.value)
        row.setText(Cols.association, rel_text)

    ### Action
    def _serial_number_entered(self):
        self.controller._serial_number = self.serial_number.text()
        ensureDeferred(self.controller.sync_now()).addErrback(open_error_dialog)

    def _debug_clicked(self):
        sync_def = ensureDeferred(self.controller.debug())
        sync_def.addErrback(open_error_dialog)

    def _sync_now_clicked(self):
        sync_def = ensureDeferred(self.controller.sync_now())
        sync_def.addErrback(open_error_dialog)

    def _sync_all_clicked(self):
        sync_def = ensureDeferred(self.controller.sync_all())
        sync_def.addErrback(open_error_dialog)

    def _login_clicked(self):
        ensureDeferred(self.controller.login_clicked()).addErrback(open_error_dialog)

    def _logout_clicked(self):
        ensureDeferred(self.controller.logout_clicked()).addErrback(open_error_dialog)

    def _serial_number_from_parameters_clicked(self):
        ensureDeferred(
            self.controller.serial_number_from_parameters_clicked()
        ).addErrback(open_error_dialog)

    def _notes_changed(self):
        changed = self.controller.notes_modified(
            self.description.text(), self.notes.toPlainText()
        )
        ensureDeferred(self._disable_notes_buttons(not changed)).addErrback(
            open_error_dialog
        )

    def _save_notes_clicked(self):
        new_desc = self.description.text()
        new_text = self.notes.toPlainText()
        ensureDeferred(
            self.controller.save_notes(self._current_file_id, new_desc, new_text)
        ).addCallback(
            lambda _: ensureDeferred(self._disable_notes_buttons(True))
        ).addErrback(
            open_error_dialog
        )

    def _reset_notes(self):
        self.notes.setPlainText(self.controller.old_notes)
        self.notes.moveCursor(QTextCursor.End)

    async def _disable_notes_buttons(self, disabled: bool):
        self.btn_save_notes.setDisabled(disabled)
        self.btn_reset_notes.setDisabled(disabled)

    def _render_context_menu(self, position: QPoint):
        item = self.files_grid.itemAt(position)
        if item is None:
            # User clicked empty space below the list items
            return
        parent = item.parent()

        menu_pos = self.files_grid.viewport().mapToGlobal(position)

        if parent is self.section_headers.raw_logs:
            self._render_raw_log_menu(menu_pos, item)
        elif parent is self.section_headers.other:
            self._render_other_file_menu(menu_pos, item)
        elif parent is self.section_headers.firmware:
            self._render_firmware_menu(menu_pos, item)
        elif parent is self.section_headers.params:
            self._render_param_file_menu(menu_pos, item)
        elif parent is self.section_headers.pmvs:
            self._render_pmvs_file_menu(menu_pos, item)

    def _render_firmware_menu(self, menu_pos: QPoint, item: QTreeWidgetItem):
        menu = QMenu(self.files_grid)
        send_to_inverter = menu.addAction("Flash to inverter")
        send_to_inverter.setDisabled(True)
        save_as = menu.addAction("Save firmware as...")

        action = menu.exec(menu_pos)

        if action is None:
            pass
        elif action is save_as:
            ensureDeferred(self.controller.save_file_as_clicked(item))
        elif action is send_to_inverter:
            pass  # TODO: [EPC] Implement this

    def _render_other_file_menu(self, menu_pos: QPoint, item: QTreeWidgetItem):
        menu = QMenu(self.files_grid)
        open = menu.addAction("Open file")
        save_as = menu.addAction("Save file as...")

        action = menu.exec(menu_pos)

        if action is None:
            pass
        elif action is open:
            self.controller.open_file(item)
        elif action is save_as:
            ensureDeferred(self.controller.save_file_as_clicked(item))

    def _render_param_file_menu(self, menu_pos: QPoint, item: QTreeWidgetItem):
        menu = QMenu(self.files_grid)
        dummy = menu.addAction("Send dummy File Loaded event to web tool")
        scratch = menu.addAction("Send to scratch")
        active = menu.addAction("Send to active")
        inverter = menu.addAction("Send to inverter")
        save_as = menu.addAction("Save file as...")

        btn: QAction
        [btn.setDisabled(True) for btn in [scratch, active, inverter]]

        action = menu.exec(menu_pos)

        if action is None:
            pass
        elif action is dummy:
            ensureDeferred(self.controller.send_dummy_param_event(item))
        elif action is scratch:
            pass
        elif action is active:
            pass
        elif action is inverter:
            pass
        elif action is save_as:
            ensureDeferred(self.controller.save_file_as_clicked(item))

    def _render_pmvs_file_menu(self, menu_pos: QPoint, item: QTreeWidgetItem):
        menu = QMenu(self.files_grid)
        save_as = menu.addAction("Save file as...")

        action = menu.exec(menu_pos)

        if action is None:
            pass
        elif action is save_as:
            ensureDeferred(self.controller.save_file_as_clicked(item))

    def _render_raw_log_menu(self, menu_pos: QPoint, item: QTreeWidgetItem):
        # item_hash = next(hash for hash, row in self.pending_log_rows.items() if row == item)
        cached = self.controller.is_file_cached_locally(item)

        menu = QMenu(self.files_grid)
        download_local = menu.addAction("Download local copy")
        process_log = menu.addAction("Process raw log")
        save_as = menu.addAction("Save file as...")
        delete_local = menu.addAction("Delete local copy")

        download_local.setDisabled(cached)
        process_log.setDisabled(True and not cached)  # Disable until implemented
        save_as.setDisabled(not cached)
        delete_local.setDisabled(not cached)

        action = menu.exec(menu_pos)

        if action is None:
            pass
        elif action is download_local:
            file_hash = self.controller.get_hash_for_row(item)
            ensureDeferred(self.controller.download_log(file_hash)).addErrback(
                open_error_dialog
            ).addCallback(lambda _: self.show_check_icon(item, Cols.local))
        elif action is process_log:
            pass
            # TODO: Implement this
        elif action is save_as:
            ensureDeferred(self.controller.save_file_as_clicked(item))
        elif action is delete_local:
            file_hash = self.controller.get_hash_for_row(item)
            self.controller.cache_manager.delete_from_cache(file_hash)
            self.show_cross_status_icon(item, Cols.local)
            pass

    def _show_sync_status_icon(
        self, row: QTreeWidgetItem, col: int, icon: str, color: QColor
    ):
        row.setText(col, icon)
        row.setForeground(col, color)

    def show_check_icon(self, row: QTreeWidgetItem, col: int):
        self._show_sync_status_icon(row, col, self.fa_check, self.color_green)

    def show_question_icon(self, row: QTreeWidgetItem, col: int):
        self._show_sync_status_icon(row, col, self.fa_question, self.color_red)

    def show_cross_status_icon(self, row: QTreeWidgetItem, col: int):
        self._show_sync_status_icon(row, col, self.fa_cross, self.color_gray)

    ### UI Update methods
    def show_logged_out_warning(self):
        error = (
            "Warning: You are not currently logged in to EPC Sync. "
            + "To sync the latest configuration files for this inverter, login here:"
        )
        self.lbl_login_status.setText(f"<font color='red'><b>{error}</b></font>")
        self._show_login_bar_widgets(False)

    def show_logged_in_status(self, connected: bool, username: str = None):
        status = "connected" if connected else "offline"
        message = f"Internet status: {status}."
        if connected:
            message = f"Logged in as {username}. " + message

        self.lbl_login_status.setText(message)
        self._show_login_bar_widgets(connected)

    def _show_login_bar_widgets(self, enabled: bool):
        self.btn_login.setHidden(enabled)
        self.btn_logout.setHidden(not enabled)

        self.lbl_serial_number.setHidden(not enabled)
        self.serial_number.setHidden(not enabled)
        self.inverter_error.setHidden(not enabled)

    def show_file_details(self, association, readonly_description=False):
        if association is None:
            # Clicked on a section header
            self.filename.clear()
            self.version.clear()
            self.controller.set_original_notes("", "")
            self.description.clear()
            self.description.setReadOnly(True)
            self.notes.clear()
            self.notes.setReadOnly(True)
            return

        self.controller.set_original_notes(
            association["file"]["description"], association["file"]["notes"]
        )

        self._current_file_id = association["file"]["id"]
        self.filename.setText(association["file"]["filename"])
        self.version.setText(association["file"]["version"])
        self.description.setText(association["file"]["description"])
        self.description.setReadOnly(readonly_description)
        self.notes.setPlainText(association["file"]["notes"])
        self.notes.setReadOnly(readonly_description)

    def show_inverter_error(self, error):
        if error is None:
            self.inverter_error.setText("")
        else:
            self.inverter_error.setText(f"<font color='red'>{error}</font>")

    def show_sync_time(self, time: datetime):
        self.lbl_last_sync.setText(f"Last sync at: {time.strftime(self.time_format)}")

    def add_log_error_line(self, message: str):
        self.add_log_line(f"<font color='#cc0000'>{message}</font>")

    def add_log_line(self, message: str):
        timestamp = datetime.now()
        new_text = f"<font color='lightGray'>[{timestamp.strftime(self.time_format).strip()}]</font> {message}<br/>"
        self._log_text = new_text + self._log_text
        self.event_log.setText(self._log_text)


# .ui files need a direct module attribute, not a class method, afaict.
FilesViewQtBuilder = FilesView.qt_build



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

# See file COPYING in this source tree
__copyright__ = "\n".join(
    [
        "Copyright 2010, Kunal Anand",
        # https://gist.github.com/327585
        "Copyright 2011, Tennessee Carmel-Veilleux",
        # http://www.tentech.ca/2011/05/stream-tee-in-python-saving-stdout-to-file-while-keeping-the-console-alive/
        "Copyright 2016, EPC Power Corp."
        # Adjusted to take a list of principals
    ]
)
__license__ = "GPLv2+"


class Tee:
    def __init__(self, principals):
        if len(principals) < 1:
            # TODO: be more specific with exception selection
            raise Exception("Tee requires at least one principal")

        self.principals = principals
        self.__missing_method_name = None  # Hack!

    def __getattribute__(self, name):
        if name == "__isabstractmethod__":
            # When upgrading to Twisted 21.2.0 there was a change to
            # twisted.python.log._GlobalStartStopObserver to make it an ABC.
            # This makes DefaultObserver also an ABC, but it stores sys.stderr
            # as a class attribute.  Thus, when creating an instance the ABC
            # metaclasses queries for the .__isabstractmethod__ attribute and
            # we respond with self.__methodmissing__ which is truthy and so
            # ABC concludes the instance of Tee is in fact an abstract method...
            # Let's not be an abstract method.  (also, let's delete this)
            return False

        return object.__getattribute__(self, name)

    def __getattr__(self, name):
        self.__missing_method_name = name  # Could also be a property
        return getattr(self, "__methodmissing__")

    def __methodmissing__(self, *args, **kwargs):
        call = getattr(self.principals[0], self.__missing_method_name, None)
        if call is not None:
            result = call(*args, **kwargs)
        else:
            result = None

        for principal in self.principals[1:]:
            call = getattr(principal, self.__missing_method_name, None)
            if call is not None:
                call(*args, **kwargs)

        return result


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import pytest


def pytest_addoption(parser):
    parser.addoption(
        "--device-present",
        action="store_true",
        default=False,
        help="Run tests that require a device be connected",
    )
    parser.addoption(
        "--run-factory",
        action="store_true",
        default=False,
        help="Run tests that require a factory device file",
    )


def pytest_collection_modifyitems(config, items):
    if not config.getoption("--device-present"):
        device_present = pytest.mark.skip(
            reason="need --device-present option to run",
        )
        for item in items:
            if "require_device" in item.keywords:
                item.add_marker(device_present)

    if not config.getoption("--run-factory"):
        factory = pytest.mark.skip(
            reason="need --run-factory option to run",
        )
        for item in items:
            if "factory" in item.keywords:
                item.add_marker(factory)



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

from PyQt5.QtCore import QEvent, Qt

import epyqlib.widgets.abstractwidget
import epyqlib.widgets.text_ui


# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class Text(epyqlib.widgets.abstractwidget.AbstractWidget):
    def __init__(self, parent=None, in_designer=False):
        self._frame = None
        self._signal = None

        super().__init__(
            ui_class=epyqlib.widgets.text_ui.Ui_Form,
            parent=parent,
            in_designer=in_designer,
        )

    def set_value(self, value):
        # TODO: quit hardcoding this and it's better implemented elsewhere
        if value is not None:
            value *= self._conversion_multiplier

        if value is None:
            value = "-"
        else:
            if self.signal_object is None:
                value = "{0:.2f}".format(value)
            elif len(self.signal_object.enumeration) > 0:
                value = self.signal_object.short_string
            else:
                decimal_places = (
                    None if self.decimal_places < 0 else self.decimal_places
                )
                value = self.signal_object.format_float(
                    value=value, decimal_places=decimal_places
                )

        self.ui.value.setText(value)

    def set_signal(self, *args, **kwargs):
        epyqlib.widgets.abstractwidget.AbstractWidget.set_signal(self, *args, **kwargs)

        self.update_layout()

    # TODO: CAMPid 097327143264214321432453216453762354
    def update_layout(self):
        width = self.calculate_max_value_width()
        if width is not None:
            self.ui.value.setMinimumWidth(width)

        if self.signal_object is not None:
            alignment = (
                Qt.AlignRight
                if len(self.signal_object.enumeration) == 0
                else Qt.AlignCenter
            )
            alignment |= Qt.AlignVCenter
            self.ui.value.setAlignment(alignment)

    def event(self, *args, **kwargs):
        result = epyqlib.widgets.abstractwidget.AbstractWidget.event(
            self, *args, **kwargs
        )

        event = args[0]
        if event.type() == QEvent.Polish:
            self.update_layout()

        return result

    # TODO: CAMPid 989849193479134917954791341
    def calculate_max_value_width(self):
        if self.signal_object is None:
            return None

        signal = self.signal_object

        decimal_places = self.decimal_places if self.decimal_places >= 0 else None

        longer = max(
            [
                signal.format_float(
                    value=v * self._conversion_multiplier, decimal_places=decimal_places
                )
                for v in [signal.min, signal.max]
            ],
            key=len,
        )

        digits = len(longer)

        if "." in longer:
            decimal = "."
            digits -= 1
        else:
            decimal = ""

        self.ui.value.setVisible(self.ui.value.isVisibleTo(self))
        metric = self.ui.value.fontMetrics()
        chars = ["{:}".format(i) for i in range(10)]
        widths = [metric.width(c) for c in chars]
        widest_width = max(widths)
        widest_char = chars[widths.index(widest_width)]
        string = "{}".format((widest_char * digits) + decimal)

        strings = signal.enumeration_strings()
        strings.append(string)

        # TODO: really figure out the spacing needed but for now
        #       just add a space on each side for buffer space
        strings = [s for s in strings]

        return max([metric.width(s) for s in strings])


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import epyqlib.pm.valuesetmodel
import epyqlib.tests.test_attrsmodel

# See file COPYING in this source tree
__copyright__ = "Copyright 2017, EPC Power Corp."
__license__ = "GPLv2+"


TestAttrsModel = epyqlib.attrsmodel.build_tests(
    types=epyqlib.pm.valuesetmodel.types,
    root_type=epyqlib.pm.valuesetmodel.Root,
    columns=epyqlib.pm.valuesetmodel.columns,
)



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import epyqlib.widgets.abstractwidget

from PyQt5.QtCore import pyqtProperty, pyqtSlot, QSizeF, QRectF, Qt
from PyQt5.QtGui import QPainter, QColor

# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


def centered_rectangle(width, height):
    return QRectF(-width / 2, -height / 2, width, height)


class LineBar(epyqlib.widgets.abstractwidget.AbstractWidget):
    def __init__(self, parent=None, in_designer=False):
        self.in_designer = in_designer
        super().__init__(parent=parent, in_designer=in_designer)

        self._background_color = QColor("#474747")
        self._color = QColor("#39C550")
        self._maximum = 100
        self._minimum = 0
        self._thickness = 10
        self._value = 75

        self._reference_value = 0
        self._reference_marker_size = QSizeF(4, 4)

        self._override_range = False

    # TODO  use actual OverrideRange mixin
    @pyqtProperty(bool)
    def override_range(self):
        return self._override_range

    @override_range.setter
    def override_range(self, override):
        self._override_range = bool(override)

    @pyqtProperty(float)
    def reference_value(self):
        return self._reference_value

    @reference_value.setter
    def reference_value(self, angle):
        self._reference_value = angle

        self.update()

    @pyqtProperty(QSizeF)
    def reference_marker_size(self):
        return self._reference_marker_size

    @reference_marker_size.setter
    def reference_marker_size(self, angle):
        self._reference_marker_size = angle

        self.update_layout()

    @pyqtProperty(float)
    def thickness(self):
        return self._thickness

    @thickness.setter
    def thickness(self, thickness):
        self._thickness = thickness

        self.update_layout()

    @pyqtProperty("QColor")
    def color(self):
        return self._color

    @color.setter
    def color(self, color):
        self._color = QColor(color)

        self.update()

    @pyqtProperty("QColor")
    def background_color(self):
        return self._background_color

    @background_color.setter
    def background_color(self, color):
        self._background_color = QColor(color)

        self.update()

    def paintEvent(self, event):
        epyqlib.widgets.abstractwidget.AbstractWidget.paintEvent(self, event)

        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing, True)
        painter.translate(self.width() / 2, self.height() / 2)

        painter.fillRect(
            centered_rectangle(self.thickness, self.height()), self.background_color
        )

        reference = QRectF(
            -self.thickness / 2, -self.height() / 2, self.thickness, self.height()
        )

        painter.fillRect(
            centered_rectangle(
                self.thickness + 2 * self.reference_marker_size.width(),
                self.reference_marker_size.height(),
            ),
            self.color,
        )

        height = (
            (self.height() / 2)
            * (self.value - self.reference_value)
            / (self.maximum - self.reference_value)
        )

        painter.fillRect(
            QRectF(-self.thickness / 2, 0, self.thickness, -height), self.color
        )

    @pyqtSlot(float)
    def set_value(self, value):
        self.value = value

    @pyqtProperty(float)
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

        self.update()

    @pyqtProperty(float)
    def maximum(self):
        return self._maximum

    @maximum.setter
    def maximum(self, maximum):
        self._maximum = maximum

        self.update()

    @pyqtProperty(float)
    def minimum(self):
        return self._minimum

    @minimum.setter
    def minimum(self, minimum):
        self._minimum = minimum

        self.update()

    # TODO: CAMPid 2397847962541678243196352195498
    def set_range(self, min=None, max=None):
        # TODO: stop using min/max
        minimum = min
        maximum = max

        if self.override_range:
            minimum = self.minimum
            maximum = self.maximum

        if minimum == maximum:
            # TODO: pick the right exception
            raise Exception("Min and max may not be the same")
        elif minimum > maximum:
            # TODO: pick the right exception
            raise Exception("Min must be less than max")

        if minimum is not None:
            self._minimum = minimum
        if maximum is not None:
            self._maximum = maximum

    def update_layout(self):
        minimum = self.thickness + 2 * self.reference_marker_size.width()
        self.setMinimumWidth(minimum)

        self.update()

    def set_unit_text(self, units):
        pass


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import collections
import inspect
import itertools
import json
import pathlib
import uuid

import attr
import graham
import pytest
from PyQt5 import QtCore

import epyqlib.attrsmodel
import epyqlib.pm.parametermodel
import epyqlib.tests.test_attrsmodel


with open(pathlib.Path(__file__).with_suffix(".json")) as f:
    serialized_sample = f.read()


@attr.s
class SampleModel:
    root = attr.ib(
        factory=lambda: epyqlib.pm.parametermodel.Root(
            uuid="ce63b990-8061-4b9f-9ad7-d2b60fe8b4c9",
        ),
    )
    model = attr.ib(default=None)
    group_a = attr.ib(default=None)
    enumerations = attr.ib(default=None)
    letters_enumeration = attr.ib(default=None)
    letters_enumerators = attr.ib(factory=dict)
    numbers_enumeration = attr.ib(default=None)
    numbers_enumerators = attr.ib(factory=dict)
    table = attr.ib(default=None)

    @classmethod
    def build(cls):
        sample_model = cls()
        sample_model.model = epyqlib.attrsmodel.Model(
            root=sample_model.root,
            columns=epyqlib.pm.parametermodel.columns,
        )

        return sample_model

    def fill(self):
        self.group_a = epyqlib.pm.parametermodel.Group(
            name="Group A",
            uuid="78fb0043-b46f-4198-86ed-3ba8d48bf4c8",
        )
        parameter_a_a = epyqlib.pm.parametermodel.Parameter(
            name="Parameter A A",
            uuid="fd184494-35a2-4b3a-970c-9a3d98cc15cd",
        )
        group_a_b = epyqlib.pm.parametermodel.Group(
            name="Group A B",
            uuid="5db07237-82c8-488f-925b-1a2a2802fb5b",
        )
        parameter_b = epyqlib.pm.parametermodel.Parameter(
            name="Parameter B",
            default=42,
            uuid="ee545382-a651-479b-9dc2-a4c3d42c230c",
        )
        group_c = epyqlib.pm.parametermodel.Group(
            name="Group C",
            uuid="7878effc-2d29-4719-bf0d-192e18d0f27c",
        )

        self.root.append_child(self.group_a)
        self.group_a.append_child(parameter_a_a)
        self.group_a.append_child(group_a_b)

        self.root.append_child(parameter_b)

        self.root.append_child(group_c)

        self.enumerations = epyqlib.pm.parametermodel.Enumerations(
            name="Enumerations",
            uuid="4f71f673-af26-4c70-8d3e-910902163e98",
        )
        self.root.append_child(self.enumerations)

        self.letters_enumeration = epyqlib.pm.parametermodel.Enumeration(
            name="Letters",
            uuid="fd1b8036-8aef-4914-9165-ed3118bd8ecd",
        )
        self.enumerations.append_child(self.letters_enumeration)
        self.letters_enumerators = {
            letter: epyqlib.pm.parametermodel.Enumerator(name=letter, uuid=uuid)
            for letter, uuid in (
                ("a", "a2cf0c0c-c55a-4a0c-b830-571b5abe089b"),
                ("b", "effdff81-7da1-4cd8-863f-44f33c75813e"),
            )
        }
        for letter, enumerator in sorted(self.letters_enumerators.items()):
            self.letters_enumeration.append_child(enumerator)

        self.numbers_enumeration = epyqlib.pm.parametermodel.Enumeration(
            name="numbers",
            uuid="86111e99-624c-499e-a560-6a77db461e43",
        )
        self.enumerations.append_child(self.numbers_enumeration)
        self.numbers_enumerators = {
            number: epyqlib.pm.parametermodel.Enumerator(
                name=number,
                uuid=uuid,
            )
            for number, uuid in (
                ("1", "165ccc7b-7ffc-4546-add2-11f0855f86d1"),
                ("2", "64e3e07c-dd4f-439a-850c-69508bfd545c"),
            )
        }
        for number, enumerator in sorted(self.numbers_enumerators.items()):
            self.numbers_enumeration.append_child(enumerator)

        self.table = epyqlib.pm.parametermodel.Table(
            name="Table A",
            uuid="58b7bf0d-4dec-4707-a976-d49a856eb3ba",
        )
        self.group_a.append_child(self.table)
        self.table.append_child(
            epyqlib.pm.parametermodel.TableEnumerationReference(
                name="Letters",
                uuid="d80162ea-bd8c-4936-bca1-732b11205423",
                enumeration_uuid=self.letters_enumeration.uuid,
            ),
        )
        self.table.append_child(
            epyqlib.pm.parametermodel.TableEnumerationReference(
                name="Numbers",
                uuid="3d98c907-0b81-4e29-be9b-2620ef44ce26",
                enumeration_uuid=self.numbers_enumeration.uuid,
            ),
        )

        array_a = epyqlib.pm.parametermodel.Array(
            name="Table Array A",
            uuid="77ed47bb-fbea-4756-b1c4-2e2be02aebc7",
        )
        array_a.append_child(
            epyqlib.pm.parametermodel.Parameter(
                name="A0",
                uuid="a32ea0c1-0838-430f-8d3c-192b5bff3047",
            )
        )
        array_a.length = 2
        array_a.children[1].name = "A1"
        array_a.children[1].uuid = uuid.UUID(
            "9d642888-b7f0-4d04-8ebf-10887249b179",
        )
        self.table.append_child(array_a)

        array_b = epyqlib.pm.parametermodel.Array(
            name="Table Array B",
            uuid="acda3b53-785c-4588-8b35-d58a9f2e10a8",
        )
        array_b.append_child(
            epyqlib.pm.parametermodel.Parameter(
                name="B0",
                uuid="1c7d0fac-aaa6-415b-b7c1-08d1f5d07097",
            )
        )
        array_b.length = 2
        array_b.children[1].name = "B1"
        array_b.children[1].uuid = uuid.UUID(
            "32666885-2934-481f-88fd-7e3157f702b1",
        )
        self.table.append_child(array_b)

        table_group_root = epyqlib.pm.parametermodel.TableGroupElement(
            name="Tree",
            uuid="428124a2-6fe3-4b40-9e23-f78e19c7aa7a",
        )

        table_group_root_a = epyqlib.pm.parametermodel.TableGroupElement(
            original=self.letters_enumerators["a"],
            uuid="904d287d-2cdd-487e-ada9-ed0f8a552078",
            path=(self.letters_enumerators["a"].uuid,),
        )
        table_group_root_a_1 = epyqlib.pm.parametermodel.TableGroupElement(
            original=self.numbers_enumerators["1"],
            uuid="9d37bb7f-2618-4ce8-b9f9-214c19851361",
            path=table_group_root_a.path + (self.numbers_enumerators["1"].uuid,),
        )
        table_group_root_a_1_a = epyqlib.pm.parametermodel.TableGroupElement(
            original=array_a,
            uuid="993ededd-6083-4c1e-aa72-bf08ee3ea61c",
            path=table_group_root_a_1.path + (array_a.uuid,),
        )
        self.add_array_elements(
            array=array_a,
            group=table_group_root_a_1_a,
            uuids=[
                "403b3a2c-f94a-490c-b112-87cdd3551e55",
                "3fd1f0e6-09e7-413e-b169-3fafff126bbc",
            ],
        )
        table_group_root_a_1.append_child(table_group_root_a_1_a)

        table_group_root_a_1_b = epyqlib.pm.parametermodel.TableGroupElement(
            original=array_b,
            uuid="72ce1cf9-0c90-48e9-a37c-aee1b36fca80",
            path=table_group_root_a_1.path + (array_b.uuid,),
        )
        self.add_array_elements(
            array=array_b,
            group=table_group_root_a_1_b,
            uuids=[
                "106cc983-0321-4ced-9e72-ab56d6bf1e7f",
                "80fd2eb7-6c89-4b65-ad35-05d63522f991",
            ],
        )
        table_group_root_a_1.append_child(table_group_root_a_1_b)
        table_group_root_a.append_child(table_group_root_a_1)
        table_group_root_a_2 = epyqlib.pm.parametermodel.TableGroupElement(
            original=self.numbers_enumerators["2"],
            uuid="b748bc78-0267-477e-95ec-e64a6686e010",
            path=table_group_root_a.path + (self.numbers_enumerators["2"].uuid,),
        )
        table_group_root_a_2_a = epyqlib.pm.parametermodel.TableGroupElement(
            original=array_a,
            uuid="712cda8a-4744-4da2-a1d1-1573625bbf14",
            path=table_group_root_a_2.path + (array_a.uuid,),
        )
        self.add_array_elements(
            array=array_a,
            group=table_group_root_a_2_a,
            uuids=[
                "290b2762-61c7-4666-8f99-af3c0113d4a3",
                "aee3a1de-b35a-4e7b-b274-1ef6766d5889",
            ],
        )
        table_group_root_a_2.append_child(table_group_root_a_2_a)
        table_group_root_a_2_b = epyqlib.pm.parametermodel.TableGroupElement(
            original=array_b,
            uuid="e74d9047-1635-43f3-825b-6844fab96a00",
            path=table_group_root_a_2.path + (array_b.uuid,),
        )
        self.add_array_elements(
            array=array_b,
            group=table_group_root_a_2_b,
            uuids=[
                "99a1be09-c4e0-41f1-bda4-d0f139038cbf",
                "fa79e1aa-647d-4783-8b16-fe752ac8e8c4",
            ],
        )
        table_group_root_a_2.append_child(table_group_root_a_2_b)
        table_group_root_a.append_child(table_group_root_a_2)
        table_group_root.append_child(table_group_root_a)

        table_group_root_b = epyqlib.pm.parametermodel.TableGroupElement(
            original=self.letters_enumerators["b"],
            uuid="104d1901-524e-412b-84e7-98707dce362d",
            path=(self.letters_enumerators["b"].uuid,),
        )
        table_group_root_b_1 = epyqlib.pm.parametermodel.TableGroupElement(
            original=self.numbers_enumerators["1"],
            uuid="9dc75463-f6a9-411f-b0a7-5c311086ff69",
            path=table_group_root_b.path + (self.numbers_enumerators["1"].uuid,),
        )
        table_group_root_b_1_a = epyqlib.pm.parametermodel.TableGroupElement(
            original=array_a,
            uuid="f0989f09-4066-4047-8d61-c134c6e1e55c",
            path=table_group_root_b_1.path + (array_a.uuid,),
        )
        self.add_array_elements(
            array=array_a,
            group=table_group_root_b_1_a,
            uuids=[
                "c076b3e1-a287-468c-80e4-43523a021feb",
                "937be4d7-b578-4daa-aef0-4065198a613d",
            ],
        )
        table_group_root_b_1.append_child(table_group_root_b_1_a)
        table_group_root_b_1_b = epyqlib.pm.parametermodel.TableGroupElement(
            original=array_b,
            uuid="18e84e98-e8ad-4d4a-9180-951350f6590c",
            path=table_group_root_b_1.path + (array_b.uuid,),
        )
        self.add_array_elements(
            array=array_b,
            group=table_group_root_b_1_b,
            uuids=[
                "286079b0-6028-4959-8396-78c626cdab46",
                "de30995a-ba26-49cf-abd0-84d3dc62b3eb",
            ],
        )
        table_group_root_b_1.append_child(table_group_root_b_1_b)
        table_group_root_b.append_child(table_group_root_b_1)
        table_group_root_b_2 = epyqlib.pm.parametermodel.TableGroupElement(
            original=self.numbers_enumerators["2"],
            uuid="1178e67b-ddda-43f3-a26a-30d0e10a2bd3",
            path=table_group_root_b.path + (self.numbers_enumerators["2"].uuid,),
        )
        table_group_root_b_2_a = epyqlib.pm.parametermodel.TableGroupElement(
            original=array_a,
            uuid="1134bc2e-93d9-4b33-8164-dca0520f3910",
            path=table_group_root_b_2.path + (array_a.uuid,),
        )
        self.add_array_elements(
            array=array_a,
            group=table_group_root_b_2_a,
            uuids=[
                "6a151510-c6e5-4673-9198-d3d9fb0dfcdc",
                "dd3887db-df84-4f5f-ae6f-98201403c1d2",
            ],
        )
        table_group_root_b_2.append_child(table_group_root_b_2_a)
        table_group_root_b_2_b = epyqlib.pm.parametermodel.TableGroupElement(
            original=array_b,
            uuid="d84963e3-164e-4df8-9f5e-cf6717e20b0d",
            path=table_group_root_b_2.path + (array_b.uuid,),
        )
        self.add_array_elements(
            array=array_b,
            group=table_group_root_b_2_b,
            uuids=[
                "d01008cc-cef0-4c95-b642-71e974690416",
                "2b16a382-2fa2-41b4-a0dc-161c7496d13a",
            ],
        )
        table_group_root_b_2.append_child(table_group_root_b_2_b)
        table_group_root_b.append_child(table_group_root_b_2)
        table_group_root.append_child(table_group_root_b)

        (automatic_table_root,) = [
            child
            for child in self.table.children
            if isinstance(child, epyqlib.pm.parametermodel.TableGroupElement)
        ]

        with self.table._ignore_children():
            self.table.remove_child(child=automatic_table_root)

            self.table.append_child(table_group_root)

        self.model.update_nodes()

    def add_array_elements(self, array, group, uuids):
        assert len(array.children) == len(uuids)

        for child, a_uuid in zip(array.children, uuids):
            element = epyqlib.pm.parametermodel.TableArrayElement(
                original=child,
                uuid=a_uuid,
                path=group.path + (child.uuid,),
            )
            group.append_child(element)


@pytest.fixture
def sample():
    sample_model = SampleModel.build()
    sample_model.fill()

    return sample_model


(default_column_index,) = (
    index
    for index, column in enumerate(epyqlib.pm.parametermodel.columns)
    if column.name == "Default"
)


@pytest.mark.parametrize(
    "column,target",
    (
        (0, "Parameter B"),
        (default_column_index, 42),
    ),
)
def test_search_in_column(sample, column, target):
    proxy = QtCore.QSortFilterProxyModel()
    proxy.setSourceModel(sample.model.model)

    (index,) = proxy.match(
        proxy.index(0, column),
        QtCore.Qt.DisplayRole,
        target,
    )


@pytest.mark.parametrize(
    "column,target",
    (
        (0, "Parameter B"),
        (default_column_index, 42),
    ),
)
def test_proxy_search_in_column(sample, column, target):
    proxy = epyqlib.utils.qt.PySortFilterProxyModel(filter_column=0)
    proxy.setSourceModel(sample.model.model)

    (index,) = proxy.match(
        proxy.index(0, column),
        QtCore.Qt.DisplayRole,
        target,
    )

    match_node = sample.model.node_from_index(proxy.mapToSource(index))

    index = proxy.search(
        text=target,
        search_from=sample.model.index_from_node(sample.model.root),
        column=column,
    )

    search_node = sample.model.node_from_index(proxy.mapToSource(index))

    assert match_node is search_node


def test_sample_dumps_consistently(sample):
    dumped = graham.dumps(sample.root, indent=4)

    def load(s):
        return json.loads(s, object_pairs_hook=collections.OrderedDict)

    print()
    print(dumped.data)

    assert load(dumped.data) == load(serialized_sample)


def test_sample_loads_consistently(sample):
    loaded = graham.schema(type(sample.root)).loads(serialized_sample)
    sample.root.model = None

    assert loaded.data == sample.root


def test_table_addition(sample):
    table = epyqlib.pm.parametermodel.Table()
    sample.group_a.append_child(table)

    seconds = epyqlib.pm.parametermodel.Array(name="seconds")
    table.append_child(seconds)
    hertz = epyqlib.pm.parametermodel.Array(name="hertz")
    table.append_child(hertz)

    letters_reference = epyqlib.pm.parametermodel.TableEnumerationReference()
    letters_reference.link(sample.letters_enumeration)
    table.append_child(letters_reference)

    numbers_reference = epyqlib.pm.parametermodel.TableEnumerationReference()
    numbers_reference.link(sample.numbers_enumeration)
    table.append_child(numbers_reference)

    # assert letters_reference.name == sample.letters_enumeration.name


def test_table_add_enumeration_with_uuid(sample):
    punctuation_enumeration = epyqlib.pm.parametermodel.Enumeration(
        name="punctuation",
    )
    sample.enumerations.append_child(punctuation_enumeration)
    punctuation_enumerators = {
        punctuation: epyqlib.pm.parametermodel.Enumerator(name=punctuation)
        for punctuation in "!@#$%^"
    }
    for punctuation, enumerator in sorted(punctuation_enumerators.items()):
        punctuation_enumeration.append_child(enumerator)

    ref = epyqlib.pm.parametermodel.TableEnumerationReference(
        enumeration_uuid=punctuation_enumeration.uuid,
    )

    sample.table.append_child(ref)


def verify_table_automatic_groups(sample):
    groups = [
        child
        for child in sample.table.children
        if isinstance(child, epyqlib.pm.parametermodel.TableGroupElement)
    ]

    assert len(groups) == 1

    (group,) = groups

    enumeration_references = [
        child
        for child in sample.table.children
        if isinstance(child, epyqlib.pm.parametermodel.TableEnumerationReference)
    ]

    enumerations = []
    for enumeration_reference in enumeration_references:
        (enumeration,) = sample.root.nodes_by_attribute(
            attribute_name="uuid",
            attribute_value=enumeration_reference.enumeration_uuid,
        )
        enumerations.append(enumeration)

    arrays = [
        child
        for child in sample.table.children
        if isinstance(child, epyqlib.pm.parametermodel.Array)
    ]

    zipped = itertools.zip_longest(
        group.children,
        enumerations[0].children,
    )
    for child, enumerator in zipped:
        print("checking: {}".format((enumerator.name)))
        assert child.name == enumerator.name

        zipped = itertools.zip_longest(
            child.children,
            enumerations[1].children,
        )
        for subchild, subenumerator in zipped:
            print("checking: {}".format((enumerator.name, subenumerator.name)))
            assert subchild.name == subenumerator.name

            zipped = itertools.zip_longest(
                subchild.children,
                arrays,
            )
            for subsubchild, array in zipped:
                print(
                    "checking: {}".format(
                        (enumerator.name, subenumerator.name, array.name)
                    )
                )
                assert subsubchild.name == array.name

                zipped = itertools.zip_longest(
                    subsubchild.children,
                    array.children,
                )
                for subsubsubchild, element in zipped:
                    print(
                        "checking: {}".format(
                            (
                                enumerator.name,
                                subenumerator.name,
                                array.name,
                                element.name,
                            )
                        )
                    )
                    assert subsubsubchild.name == element.name


def test_table_automatic_groups(sample):
    verify_table_automatic_groups(sample)


def test_table_automatic_groups_add_another(sample):
    array_c = epyqlib.pm.parametermodel.Array(
        name="Table Array A",
        uuid="eb812b66-f98b-4e8f-a7e8-4f76c853e437",
    )

    sample.table.append_child(array_c)

    verify_table_automatic_groups(sample)


def test_table_update_same_uuid(qapp):
    root = (
        graham.schema(epyqlib.pm.parametermodel.Root)
        .loads(
            serialized_sample,
        )
        .data
    )
    model = epyqlib.attrsmodel.Model(
        root=root,
        columns=epyqlib.pm.parametermodel.columns,
    )
    root.model = model

    (table_a,) = root.nodes_by_attribute(
        attribute_value="Table A",
        attribute_name="name",
    )

    # TODO: CAMPid 9784566547216435136479765479163496731
    def collect(node):
        def collect(node, payload):
            payload[node.uuid] = node.name

        results = {}

        node.traverse(
            call_this=collect,
            internal_nodes=True,
            payload=results,
        )

        return results

    table_a.update()

    original = collect(table_a)

    table_a.update()

    after_update = collect(table_a)

    assert after_update == original


def test_array_addable_types():
    array = epyqlib.pm.parametermodel.Array()

    value_types = (
        epyqlib.pm.parametermodel.Parameter,
        epyqlib.pm.parametermodel.Group,
    )

    assert array.addable_types() == epyqlib.attrsmodel.create_addable_types(
        value_types,
    )

    for value_type in value_types:
        child = value_type()
        array.append_child(child)

        assert array.addable_types() == epyqlib.attrsmodel.create_addable_types(
            (),
        )

        array.remove_child(child=child)

        assert array.addable_types() == epyqlib.attrsmodel.create_addable_types(
            value_types,
        )


def test_array_update_children_length():
    array = epyqlib.pm.parametermodel.Array()
    parameter = epyqlib.pm.parametermodel.Parameter()

    assert len(array.children) == 0

    array.append_child(parameter)

    assert len(array.children) == 1

    for n in (3, 7, 4, 1, 5):
        array.length = n
        assert len(array.children) == n


def test_move_enumeration_in_table(sample):
    uuids_before = {
        node.uuid for node in sample.table.nodes_by_filter(lambda node: True)
    }

    table_index = sample.model.index_from_node(sample.table)
    names_before = epyqlib.utils.qt.child_text_list_from_index(
        model=sample.model.model,
        index=table_index,
        recurse=False,
    )

    sample.table.update()
    uuids_after = {
        node.uuid for node in sample.table.nodes_by_filter(lambda node: True)
    }

    assert len(uuids_after) == len(uuids_before)
    assert uuids_after == uuids_before

    node_to_move = sample.table.children[1]

    mime_data = sample.model.mimeData(
        (sample.model.index_from_node(node_to_move),),
    )

    sample.model.dropMimeData(
        data=mime_data,
        action=QtCore.Qt.MoveAction,
        row=-1,
        column=0,
        parent=sample.model.index_from_node(sample.table),
    )

    sample.table.update()
    uuids_after = {
        node.uuid for node in sample.table.nodes_by_filter(lambda node: True)
    }

    assert len(uuids_after) == len(uuids_before)
    assert uuids_after == uuids_before

    names_after = epyqlib.utils.qt.child_text_list_from_index(
        model=sample.model.model,
        index=table_index,
        recurse=False,
    )

    assert set(names_before) == set(names_after)
    assert len(names_before) == len(names_after)
    assert names_before != names_after


TestAttrsModel = epyqlib.attrsmodel.build_tests(
    types=epyqlib.pm.parametermodel.types,
    root_type=epyqlib.pm.parametermodel.Root,
    columns=epyqlib.pm.parametermodel.columns,
)



import contextlib
import functools
import locale
import os
import pathlib
import sys

import attr


library_path = pathlib.Path(__file__).parents[2].resolve()

project_path = library_path.parents[1]

scripts_path = project_path / ".venv" / "Scripts"

examples_path = library_path / "examples" / "develop"

device_path = examples_path / "devices"

default_parameters_path = examples_path / "defaults.pmvs"
small_parameters_path = examples_path / "small.pmvs"

devices = {
    "customer": pathlib.Path("customer") / "distributed_generation.epc",
    "factory": pathlib.Path("factory") / "distributed_generation_factory.epc",
}

devices = {k: device_path / v for k, v in devices.items()}

symbol_files = {
    "customer": pathlib.Path("customer") / "EPC_DG_ID247.sym",
    "factory": pathlib.Path("factory") / "EPC_DG_ID247_FACTORY.sym",
}

symbol_files = {k: device_path / v for k, v in symbol_files.items()}

hierarchy_files = {
    "customer": pathlib.Path("customer") / "EPC_DG_ID247.parameters.json",
    "factory": pathlib.Path("factory") / "EPC_DG_ID247_FACTORY.parameters.json",
}

hierarchy_files = {k: device_path / v for k, v in hierarchy_files.items()}


def single(x):
    (y,) = x

    return y


@attr.s
class DeviceFiles:
    can = attr.ib()
    hierarchy = attr.ib()
    device = attr.ib()
    epp = attr.ib()
    pmvs = attr.ib()

    @classmethod
    def build(cls, base, version, level):
        version = pathlib.Path(base) / version
        path = version / "devices" / level

        print(path, list(path.glob("*")))

        return cls(
            can=single(path.glob("*.sym")),
            hierarchy=single(path.glob("*.json")),
            device=single(path.glob("*.epc")),
            epp=version / "small.epp",
            pmvs=version / "small.pmvs",
        )


new_examples_path = library_path / "examples"


@functools.lru_cache()
def new_devices():
    return {
        (version, level): DeviceFiles.build(
            base=new_examples_path,
            version=version,
            level=level,
        )
        for version in (
            "develop",
            "v1.2.5",
        )
        for level in (
            "customer",
            "factory",
        )
        if (version, level) not in (("v1.2.5", "customer"),)
    }


@attr.s
class Values:
    initial = attr.ib()
    input = attr.ib()
    expected = attr.ib()
    collected = attr.ib(default=attr.Factory(list))

    def collect(self, value):
        self.collected.append(value)

    def check(self):
        return all(x == y for x, y in zip(self.expected, self.collected))


@contextlib.contextmanager
def use_locale(*s):
    if len(s) == 0:
        s = ("",)

    # use setlocale() to get present settings because reasons:
    # https://docs.python.org/3/library/locale.html#locale.getlocale
    # https://docs.python.org/3/library/locale.html#locale.setlocale
    old = locale.setlocale(locale.LC_ALL)

    for name in s:
        try:
            locale.setlocale(locale.LC_ALL, name)
        except locale.Error:
            continue

        break
    else:
        assert False, "Unable to set locale to any of {}".format(s)

    yield

    locale.setlocale(locale.LC_ALL, old)



import collections
import csv
import functools
import io
import textwrap

import attr
import twisted.internet.defer
import twisted.internet.task

from PyQt5 import QtCore, QtWidgets

import epyqlib.twisted.busproxy
import epyqlib.twisted.cancalibrationprotocol as ccp
import epyqlib.twisted.nvs
import epyqlib.utils.qt
from epyqlib.tabs.files.log_manager import LogManager

__copyright__ = "Copyright 2017, EPC Power Corp."
__license__ = "GPLv2+"


class UnsupportedError(Exception):
    pass


@attr.s
class DataLogger:
    nvs = attr.ib()
    bus = attr.ib()
    device = attr.ib()
    progress = attr.ib(default=attr.Factory(epyqlib.utils.qt.Progress))
    ccp_protocol = attr.ib(init=False)
    tx_id = attr.ib(default=0x1FFFFFFF)
    rx_id = attr.ib(default=0x1FFFFFF7)

    def __attrs_post_init__(self):
        signal = self.nvs.neo.signal_by_path("CCP", "Connect", "CommandCounter")
        self.ccp_protocol = ccp.Handler(
            endianness="little" if signal.little_endian else "big",
            tx_id=self.tx_id,
            rx_id=self.rx_id,
        )
        from twisted.internet import reactor

        self.ccp_transport = epyqlib.twisted.busproxy.BusProxy(
            protocol=self.ccp_protocol, reactor=reactor, bus=self.bus
        )

        self.nv_protocol = epyqlib.twisted.nvs.Protocol()
        self.nv_transport = epyqlib.twisted.busproxy.BusProxy(
            protocol=self.nv_protocol, reactor=reactor, bus=self.bus
        )

    def pull_raw_log(self, path):
        pull_fake_log = False
        if pull_fake_log:
            d = twisted.internet.defer.execute(self._pull_raw_log_fake)
        else:
            d = self._pull_raw_log()
        d.addCallback(write_to_file, path=path)
        d.addCallback(
            lambda _: twisted.internet.defer.ensureDeferred(
                self._notify_new_raw_log(path)
            )
        )

        d.addErrback(epyqlib.utils.twisted.detour_result, self.progress.fail)
        d.addErrback(epyqlib.utils.twisted.errbackhook)

        return d

    ## NOMERGE: Remove this before merging to master
    # usage: d = twisted.internet.defer.execute(self._pull_raw_log_fake)
    def _pull_raw_log_fake(self):
        self.progress.complete(message=None)
        from time import time

        return bytes(str(time()), "utf-8")

    async def _notify_new_raw_log(self, path: str):
        build_hash_signal = self.nvs.signal_from_names("SoftwareHash", "SoftwareHash")
        serial_number_signal = self.nvs.signal_from_names("SN", "SerialNumber")

        build_hash = await self.nv_protocol.read(
            nv_signal=build_hash_signal,
            meta=epyqlib.nv.MetaEnum.value,
        )
        build_hash_integer = int(build_hash[0])
        build_hash = f"{build_hash_integer:07x}"
        serial_number = await self.nv_protocol.read(
            nv_signal=serial_number_signal,
            meta=epyqlib.nv.MetaEnum.value,
        )
        serial_number = str(serial_number[0])

        await LogManager.get_instance().add_pending_log(path, build_hash, serial_number)

    @twisted.internet.defer.inlineCallbacks
    def _pull_raw_log(self):
        unsupported = UnsupportedError(
            "Pull of raw log is not supported for this device",
        )

        try:
            recording_signal = self.nvs.signal_from_names(
                "DataloggerStatus", "DataloggerRecording"
            )
        except epyqlib.nv.NotFoundError as e:
            raise unsupported from e

        try:
            readable_octets_signal = self.nvs.signal_from_names(
                "LoggerStatus01", "ReadableOctets"
            )
        except epyqlib.nv.NotFoundError as e:
            raise unsupported from e

        from twisted.internet import reactor

        while True:
            recording = yield self.nv_protocol.read(
                nv_signal=recording_signal,
                meta=epyqlib.nv.MetaEnum.value,
            )
            recording = int(recording[0])

            if not recording:
                break

            yield twisted.internet.task.deferLater(
                clock=reactor,
                delay=0.2,
                callable=lambda: None,
            )

        readable_octets = yield self.nv_protocol.read(
            nv_signal=readable_octets_signal,
            meta=epyqlib.nv.MetaEnum.value,
        )
        readable_octets = int(readable_octets[0])

        self.progress.configure(maximum=readable_octets)

        # TODO: hardcoded station address, tsk-tsk
        yield self.ccp_protocol.connect(station_address=0)
        data = yield self.ccp_protocol.upload_block(
            address_extension=ccp.AddressExtension.data_logger,
            address=0,
            octets=readable_octets,
            progress=self.progress,
        )
        yield self.ccp_protocol.disconnect(end_of_session=1)

        seconds = self.progress.elapsed()

        completed_format = textwrap.dedent(
            """\
        Log successfully pulled

        Data time: {seconds:.3f} seconds for {bytes} bytes or {bps:.0f} bytes/second"""
        )
        message = completed_format.format(
            seconds=seconds, bytes=readable_octets, bps=readable_octets / seconds
        )

        self.progress.complete(message=message)

        twisted.internet.defer.returnValue(data)


def write_to_file(data, path):
    with open(path, "wb") as f:
        f.write(data)


def pull_raw_log(device, bus=None, parent=None):
    if bus is None:
        bus = device.bus

    filters = [("Raw", ["raw"]), ("All Files", ["*"])]
    filename = epyqlib.utils.qt.file_dialog(filters, save=True, parent=parent)

    # TODO: perhaps an exception for cancelation?  let caller ignore it?
    if filename is None:
        d = twisted.internet.defer.Deferred()
        d.callback(None)
        return d

    progress = epyqlib.utils.qt.progress_dialog(parent=device.ui)

    logger = epyqlib.datalogger.DataLogger(
        nvs=device.nvs,
        bus=bus,
        device=device,
        tx_id=device.neo_frames.frame_by_name("CCP").id,
        rx_id=device.neo_frames.frame_by_name("CCPResponse").id,
    )

    logger.progress.connect(
        progress=progress,
        label_text=("Pulling log...\n\n" + logger.progress.default_progress_label),
    )
    return logger.pull_raw_log(path=filename)


def generate_records(
    cache,
    chunks,
    data_stream,
    variables_and_chunks,
    sample_period_us,
    raw_chunks,
):
    chunk_list = list(chunks)

    scaling_cache = {}
    timestamp = 0
    while len(data_stream.read(1)) == 1:
        data_stream.seek(-1, io.SEEK_CUR)

        row = collections.OrderedDict()
        row[".time"] = timestamp / 1000000
        timestamp += sample_period_us

        def update(data, variable, scaling_cache):
            path = ".".join(variable.path())
            value = variable.variable.unpack(data)
            type_ = variable.fields.type
            scaling = 1
            if type_ in scaling_cache:
                scaling = scaling_cache[type_]
            else:
                if type_.startswith("_iq"):
                    n = type_.lstrip("_iq")
                    if n == "":
                        n = 24
                    else:
                        n = int(n)
                    scaling = 1 << n
                scaling_cache[type_] = scaling

            row[path] = value / scaling

        for variable, chunk in variables_and_chunks.items():
            partial = functools.partial(
                update, variable=variable, scaling_cache=scaling_cache
            )
            cache.subscribe(partial, chunk)

        for chunk in raw_chunks:
            chunk_bytes = bytearray(data_stream.read(len(chunk)))
            if len(chunk_bytes) != len(chunk):
                text = (
                    "Unexpected EOF found in the middle of a record.  "
                    "Continuing with partially extracted log."
                )
                raise EOFError(text)

            chunk.set_bytes(chunk_bytes)
            cache.update(chunk)

        cache.unsubscribe_all()
        yield row


def parse_log(
    cache,
    chunks,
    csv_path,
    data_stream,
    variables_and_chunks,
    sample_period_us,
    raw_chunks,
):
    with open(csv_path, "w", newline="") as f:
        writer = None

        records = generate_records(
            cache=cache,
            chunks=chunks,
            data_stream=data_stream,
            variables_and_chunks=variables_and_chunks,
            sample_period_us=sample_period_us,
            raw_chunks=raw_chunks,
        )

        for row in records:
            if writer is None:
                writer = csv.DictWriter(
                    f, fieldnames=sorted(row.keys(), key=str.casefold)
                )
                writer.writeheader()

            writer.writerow(row)



import epyqlib.scriptingview
import epyqlib.abstractpluginclass

# See file COPYING in this source tree
__copyright__ = "Copyright 2017, EPC Power Corp."
__license__ = "GPLv2+"


class ScriptingViewPlugin(epyqlib.abstractpluginclass.AbstractPlugin):
    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self._group = "EPC - General"
        self._init = epyqlib.scriptingview.ScriptingView
        self._module_path = "epyqlib.scriptingview"
        self._name = "ScriptingView"



import pathlib

import pytest

import epyqlib.cmemoryparser


here = pathlib.Path(__file__).parent

outs = tuple(sorted((here / "outs").glob("*.out")))


@pytest.mark.parametrize(
    argnames="path",
    argvalues=outs,
    ids=[out.name for out in outs],
)
def test_load(path):
    epyqlib.cmemoryparser.process_file(filename=path)



import functools
import logging

import click

# See file COPYING in this source tree
__copyright__ = "Copyright 2017, EPC Power Corp."
__license__ = "GPLv2+"


def verbose_option(f):
    @click.option("-v", count=True)
    @functools.wraps(f)
    def d(*args, v, **kwargs):
        level = logging.getLogger().getEffectiveLevel()
        step = logging.INFO - logging.DEBUG
        level -= v * step

        logging.getLogger().setLevel(level)

        f(*args, **kwargs)

    return d



#!/usr/bin/env python3

# TODO: get some docstrings in here!

# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


application = """
QWidget {{
    font-size: {base_font_size_px}px;
    qproperty-focusPolicy: NoFocus;
    color: {foreground};
}}

QWidget#MainForm {{
    background-color: {background};
}}

QWidget[fontawesome=false] {{
    font-family: Metropolis;
}}

QWidget[fontawesome=true] {{
    font-size: 36px;
    icon-size: 36px
}}

Epc {{
    qproperty-show_enumeration_value: false;
}}

QAbstractScrollArea {{
    qproperty-frameShape: NoFrame;
}}

QPushButton {{
    font-size: {base_font_size_px}px;
    min-width: 40px;
    min-height: 40px;
    height: 40px;
    color: black;
}}

QPushButton[fontawesome=true] {{
    min-width: 46px;
    width: 46px;
    max-width: 46px;
    min-height: 46px;
    height: 46px;
    max-height: 46px;
}}

QFrame {{
    qproperty-frameShadow: Plain;
}}

QLineEdit, QPushButton {{
    border-radius: 10px;
    border-width: 0px;
    border-style: solid;
}}

QLineEdit {{
    qproperty-focusPolicy: NoFocus;
    background-color: {background_blue};
}}

QPushButton:enabled {{
    background-color: {green};
}}

QPushButton:!enabled {{
    background-color: {gray};
}}

QPushButton[active=true] {{
    background: {blue};
}}

QLineEdit {{
    qproperty-frame: false;
}}

QLineEdit:!enabled {{
    background-color: {gray};
}}

QLineEdit:enabled {{
    padding: 0 8px;
    selection-background-color: darkgray;
}}

QSlider {{
    min-height: 40px;
    min-width: 30px;
}}

QSlider::groove {{
    width: 4px;
    border-radius: 2px;
    background-color: {gray};
}}

QSlider::handle {{
    height: 10px;
    border-radius: 3px;
    margin: 0 -8px;
    background-color: {blue};
}}
"""

small = """
QWidget[fontawesome=false] {{
    font-size: 15px;
}}

QPushButton[fontawesome=false] {{
    min-height: 25px;
}}

QLineEdit, QPushButton {{
    border-radius: 5px;
}}
"""

if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

from collections import OrderedDict

from PyQt5 import QtWidgets
from PyQt5.QtCore import pyqtSignal, pyqtProperty, QEvent

import epyqlib.form
import epyqlib.utils.qt


# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


factors = OrderedDict(
    [("M", 6), ("k", 3), ("h", 2), ("da", 1), ("d", -1), ("c", -2), ("m", -3), ("", 0)]
)


def conversion_multiplier(old, new):
    old_factor = None
    new_factor = None

    for factor in factors:
        if old_factor is None and old.startswith(factor):
            old_factor = factor
        if new_factor is None and new.startswith(factor):
            new_factor = factor

    old_unit = old[len(old_factor) :]
    new_unit = new[len(new_factor) :]

    if old_unit != new_unit:
        raise Exception("Units do not match: old {}, new {}".format(old, new))

    return 10 ** (factors[old_factor] - factors[new_factor])


event_type_to_name = {
    getattr(QEvent, t): t
    for t in dir(QEvent)
    if isinstance(getattr(QEvent, t), QEvent.Type)
}


def signal_path_to_string(path):
    return ";".join(s.strip() for s in path).rstrip(";")


def string_to_signal_path(s):
    return [se.strip() for se in s.split(";")]


class AbstractWidget(QtWidgets.QWidget):
    trigger_action = pyqtSignal()

    def __init__(self, ui_class=None, parent=None, in_designer=False):
        self.in_designer = in_designer
        super().__init__(parent=parent)

        self.ui = None

        self.left = None
        self.right = None

        self.signal_object = None

        self._label_override = ""
        self._tool_tip_override = ""
        self._override_units = ""

        self._conversion_multiplier = 1
        self._decimal_places = -1

        self._signal_path = [""] * 3
        self._display_units = ""
        self._action = ""

        if ui_class is None:
            self.ui = None
        else:
            self.ui = ui_class()
            self.ui.setupUi(self)

        self.set_signal(force_update=True)

        self.ignore = False

    @pyqtProperty(str)
    def action(self):
        return self._action

    @action.setter
    def action(self, action):
        self._action = action

    @pyqtProperty(str)
    def display_units(self):
        return self._display_units

    @display_units.setter
    def display_units(self, units):
        self._display_units = units
        self.update_metadata()

    def changeEvent(self, event):
        QtWidgets.QWidget.changeEvent(self, event)
        if event.type() == QEvent.ParentChange:
            self.update_metadata()

    def update_metadata(self):
        if not self.in_designer:
            return

        from PyQt5.QtDesigner import QDesignerFormWindowInterface

        parent = self

        self.set_signal(force_update=True)

        expected_type = epyqlib.form.EpcForm
        while parent is not None:
            if isinstance(parent, expected_type):
                parent.update_widget(self)
                break
            elif isinstance(parent, QDesignerFormWindowInterface):
                # Probably getting deleted so we can ignore
                break
            else:
                parent = parent.parent()
        else:
            raise Exception(
                "No valid {} widget found while searching parents".format(
                    "epyqlib.form.EpcForm"  # expected_type.__class__.__name__
                )
            )

    def set_signal_path(self, path):
        if len(path) > len(self._signal_path):
            raise Exception(
                "Passed path has length {} which is longer than the supported "
                "limit of {}: {}".format(
                    len(path),
                    len(self._signal_path),
                    path,
                )
            )

        self._signal_path = list(path) + [""] * (len(self._signal_path) - len(path))

    @pyqtProperty("QString")
    def signal_path(self):
        return signal_path_to_string(self._signal_path)

    @signal_path.setter
    def signal_path(self, value):
        self.set_signal_path(string_to_signal_path(value))
        self.update_metadata()

    @pyqtProperty(int)
    def decimal_places(self):
        return self._decimal_places

    @decimal_places.setter
    def decimal_places(self, decimal_places):
        self._decimal_places = decimal_places
        self.update_metadata()

    @pyqtProperty("QString")
    def label_override(self):
        return self._label_override

    @label_override.setter
    def label_override(self, new_label_override):
        expanded = str(new_label_override)
        expanded = expanded.encode("utf8").decode("unicode_escape")
        self._label_override = expanded
        if self.ui is not None:
            self.ui.label.setText(self.label_override)
        self.update_metadata()

    @pyqtProperty("QString")
    def tool_tip_override(self):
        return self._tool_tip_override

    @tool_tip_override.setter
    def tool_tip_override(self, new_tool_tip_override):
        self._tool_tip_override = str(new_tool_tip_override)
        self.update_tool_tip()
        self.update_metadata()

    @pyqtProperty(bool)
    def label_visible(self):
        if self.ui is not None:
            return self.ui.label.isVisibleTo(self.parent())

        return False

    @label_visible.setter
    def label_visible(self, new_visible):
        if self.ui is not None:
            self.ui.label.setVisible(new_visible)
            self.update_metadata()

    @pyqtProperty(bool)
    def units_visible(self):
        if self.ui is not None:
            if hasattr(self.ui, "units"):
                return self.ui.units.isVisibleTo(self.parent())

        return False

    @units_visible.setter
    def units_visible(self, new_visible):
        if self.ui is not None:
            if hasattr(self.ui, "units"):
                self.ui.units.setVisible(new_visible)
                self.update_metadata()

    def containing_layout(self):
        parent = self.parent()

        if parent is None:
            layout = None
        else:
            for layout in parent.findChildren(QtWidgets.QGridLayout):
                index = layout.indexOf(self)
                if index != -1:
                    break
            else:
                layout = None

        if layout == None:
            index = None

        return layout, index

    # TODO: CAMPid 943989817913241236127998452684328
    def set_label(self, new_signal=None):
        if self.ui is not None:
            if len(self.label_override) > 0:
                override = self.label_override
            else:
                override = None

            custom = self.set_label_custom(new_signal=new_signal)

            label_choices = [override, custom]

            if new_signal is not None:
                label_choices.extend([new_signal.long_name, new_signal.name])

            label_choices.append("-")

            label = next(l for l in label_choices if l is not None)

            self.ui.label.setText(label)

            if not self.label_visible:
                layout, index = self.containing_layout()

                # TODO: CAMPid 938914912312674213467977981547743
                if index is not None:
                    row, column, row_span, column_span = layout.getItemPosition(index)
                    # TODO: search in case of colspan
                    left = None
                    for offset in range(1, column + 1):
                        left_column = column - offset
                        left_layout_item = layout.itemAtPosition(row, left_column)

                        if left_layout_item is not None:
                            left_temp = left_layout_item.widget()
                            left_index = layout.indexOf(left_temp)
                            _, _, _, column_span = layout.getItemPosition(left_index)

                            if left_temp is not None:
                                if column_span < offset:
                                    break

                                left = left_temp

                    if isinstance(left, QtWidgets.QLabel):
                        left.setText(label)
                        self.left = left

    def set_label_custom(self, new_signal=None):
        return None

    @pyqtProperty(str)
    def override_units(self):
        return self._override_units

    @override_units.setter
    def override_units(self, units):
        self._override_units = units

        # TODO: CAMPid 0932498324014012080143014320
        if self.signal_object is not None and len(self.override_units) > 0:
            self._conversion_multiplier = conversion_multiplier(
                old=self.signal_object.unit, new=self.override_units
            )

    def set_units(self, units):
        if units is None:
            units = "-"
        if len(self.override_units) > 0:
            units = self.override_units

        self.set_unit_text(units)

        if self.ui is not None:
            if not self.units_visible:
                layout, index = self.containing_layout()

                # TODO: CAMPid 938914912312674213467977981547743
                if index is not None:
                    row, column, row_span, column_span = layout.getItemPosition(index)
                    right = layout.itemAtPosition(row, column + column_span)
                    if right is not None:
                        right = right.widget()
                        if isinstance(right, QtWidgets.QLabel):
                            right.setText(units)
                            self.right = right

    def set_unit_text(self, units):
        if self.ui is not None:
            self.ui.units.setText(units)

    def set_full_string(self, string):
        pass

    def set_range(self, min=None, max=None):
        pass

    def update_connection(self, signal=None):
        if signal is not self.signal_object:
            if self.signal_object is not None:
                self.signal_object.value_changed.disconnect(self.meta_set_value)

            if signal is not None:
                signal.value_changed.connect(self.meta_set_value)

            # TODO: CAMPid 0932498324014012080143014320
            if signal is not None and len(self.override_units) > 0:
                self._conversion_multiplier = conversion_multiplier(
                    old=signal.unit, new=self.override_units
                )
            else:
                self._conversion_multiplier = 1

    def meta_set_value(self, value):
        self.set_value(value)

    def set_signal(self, signal=None, force_update=False):
        if signal is not self.signal_object or force_update:
            self.set_label(new_signal=signal)
            if signal is not None:
                self.set_units(signal.unit)
                self.set_value(None)
            else:
                self.set_units(None)

            self.update_tool_tip(new_signal=signal)

            self.update_connection(signal)
            self.signal_object = signal

            if signal is not None:
                self.set_range(min=signal.min, max=signal.max)

                signal.force_value_changed()

    def update_tool_tip(self, new_signal=None):
        signal = self.signal_object if new_signal is None else new_signal

        if len(self.tool_tip_override) > 0:
            tip = self.tool_tip_override
        elif signal is not None:
            tip = signal.comment
        else:
            tip = ""

        elements = []

        if signal is not None:
            name = signal.long_name
            if name is None:
                name = signal.name
            if name is not None:
                elements.append("Name: {}".format(name))

        elements.append("Description: {}".format(tip))
        if hasattr(self, "_signal_path"):
            elements.append("Path: {}".format(" : ".join(self._signal_path)))
        contents = "<br><br>".join(elements)
        complete = '<div align="left">{}</div>'.format(contents)
        self.setToolTip(complete)
        for widget in [self.left, self.right]:
            if widget is not None:
                widget.setToolTip(complete)


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



from PyQt5 import QtCore, QtWidgets

# See file COPYING in this source tree
__copyright__ = "Copyright 2017, EPC Power Corp."
__license__ = "GPLv2+"


class TreeView(QtWidgets.QTreeView):
    def __init__(self, *args, **kwargs):
        kwargs.pop("in_designer", None)
        super().__init__(*args, **kwargs)

        self.row_columns = set()
        self.no_update_columns = set()

    def selectionCommand(self, index, event):
        column = index.column()
        is_row_column = column in self.row_columns

        if column in self.no_update_columns and not is_row_column:
            return QtCore.QItemSelectionModel.NoUpdate

        result = super().selectionCommand(index, event)

        if is_row_column:
            result |= QtCore.QItemSelectionModel.Rows

        return result



import tempfile
import os
import time

import pytest

from epyqlib.tabs.files.sync_config import SyncConfig, ConfigurationError, Vars

tempdir = tempfile.gettempdir()
filename = f"test-{int(time.time()) % 100000}.json"
full_path = os.path.join(tempdir, filename)


def load_configuration():
    print(f"Using temp dir {full_path}")
    return SyncConfig(tempdir, filename)


def cleanup():
    os.unlink(full_path)


def test_read():
    configuration = load_configuration()
    assert configuration.get("username") is None
    configuration.set("username", "123")
    assert configuration.get("username") == "123"

    configuration = load_configuration()
    assert configuration.get("username") == "123"

    cleanup()


def test_invalid_json():
    with open(full_path, "w") as file:
        file.write("[1,2,3]")
    with pytest.raises(ConfigurationError):
        load_configuration()
    cleanup()



#!/usr/bin/env python3

# TODO: get some docstrings in here!

import contextlib
import logging
import sys
import time
import typing

import attr
import can
import can.interfaces.pcan
from epyqlib.canneo import QtCanListener
import epyqlib.utils.qt
from PyQt5 import QtCore
from PyQt5.QtWidgets import QApplication

# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


@attr.s(auto_attribs=True)
class BusSettings:
    type: str
    channel: str
    bitrate: int

    @classmethod
    def from_bus(cls, real_bus):
        available_bus_types = {
            getattr(sys.modules[module_string], name): id
            for id, [module_string, name] in can.interface.BACKENDS.items()
            if module_string in sys.modules
        }

        for attribute_name in ["channel", "channel_info"]:
            channel = getattr(real_bus, attribute_name, None)
            if channel is not None:
                break
        else:
            raise Exception(f"{real_bus}")

        return cls(
            type=available_bus_types[type(real_bus)], channel=channel, bitrate=500_000
        )

    def create_bus(self):
        real_bus = can.interface.Bus(
            bustype=self.type,
            channel=self.channel,
            bitrate=self.bitrate,
        )

        return real_bus


class BusProxy:
    went_offline = epyqlib.utils.qt.Signal()

    def __init__(
        self, bus=None, timeout=0.1, transmit=True, filters=None, auto_disconnect=True
    ):
        self.filters = filters
        self.auto_disconnect = auto_disconnect

        self.timeout = timeout
        self.notifier = NotifierProxy(self)
        self.real_notifier = None
        self.tx_notifier = NotifierProxy(None)
        self.bus: typing.Optional[BusSettings] = None
        self.bus_settings = None
        self.set_bus(bus)

        self._transmit = transmit

        self.flash_timer = QtCore.QTimer()
        self.flash_timer.setSingleShot(True)
        self.flash_timer.setInterval(10 * 1000)
        self.flash_timer.timeout.connect(self.stop_flashing)

    @contextlib.contextmanager
    def managed_real_bus(self, bustype, channel):
        real_bus = can.interface.Bus(bustype=bustype, channel=channel)
        try:
            self.set_bus(bus=real_bus)
            try:
                yield
            finally:
                self.terminate()
        finally:
            real_bus.shutdown()

    @property
    def transmit(self):
        return self._transmit

    @transmit.setter
    def transmit(self, transmit):
        self._transmit = transmit

    def send(self, msg, on_success=None):
        return self._send(msg, on_success=on_success)

    def send_passive(self, msg, on_success=None):
        return self._send(msg, on_success=on_success, passive=True)

    def _send(self, msg, on_success=None, passive=False):
        if self.bus is not None and (self._transmit or passive):
            # TODO: this (the silly sleep) is really hacky and shouldn't be needed but it seems
            #       to be to keep from forcing socketcan offbus.  the issue
            #       can be recreated with the following snippet.
            # import can
            # import time
            # bus = can.interface.Bus(bustype='socketcan', channel='can0')
            # msg = can.message.Message(arbitration_id=0x00FFAB80, bytearray([0, 0, 0, 0, 0, 0, 0, 0]))
            # for i in range(50):
            #   bus.send(msg)
            #   time.sleep(.0003)
            #
            #       which results in stuff like
            #
            # altendky@tp:/epc/bin$ can0; candump -L -x can0,#FFFFFFFF | grep -E '(0[04]FFAB(88|90|80)|can0 2)'
            # (1469135699.755374) can0 00FFAB80#0000000000000000
            # (1469135699.755462) can0 00FFAB80#0000000000000000
            # (1469135699.755535) can0 00FFAB80#0000000000000000
            # (1469135699.755798) can0 00FFAB80#0000000000000000
            # (1469135699.755958) can0 00FFAB80#0000000000000000
            # (1469135699.756132) can0 00FFAB80#0000000000000000
            # (1469135699.756446) can0 00FFAB80#0000000000000000
            # (1469135699.756589) can0 20000004#000C000000000000
            # (1469135699.756589) can0 20000004#0030000000000000
            # (1469135699.756731) can0 00FFAB80#0000000000000000
            # (1469135699.757004) can0 00FFAB80#0000000000000000
            # (1469135699.757187) can0 00FFAB80#0000000000000000
            # (1469135699.757308) can0 20000040#0000000000000000
            # (1469135699.757460) can0 00FFAB80#0000000000000000
            # (1469135699.757634) can0 00FFAB80#0000000000000000
            # (1469135699.757811) can0 00FFAB80#0000000000000000
            # (1469135699.757980) can0 00FFAB80#0000000000000000
            # (1469135699.758173) can0 00FFAB80#0000000000000000
            # (1469135699.758319) can0 00FFAB80#0000000000000000
            # (1469135699.758392) can0 00FFAB80#0000000000000000
            # (1469135699.758656) can0 00FFAB80#0000000000000000
            # (1469135699.758726) can0 00FFAB80#0000000000000000
            # (1469135699.758894) can0 00FFAB80#0000000000000000

            if isinstance(self.bus, can.BusABC):
                # TODO: this is a hack to allow detection of transmitted
                #       messages later
                msg.timestamp = None

                try:
                    # TODO: I would use message=message (or msg=msg) but:
                    #       https://bitbucket.org/hardbyte/python-can/issues/52/inconsistent-send-signatures
                    self.bus.send(msg)
                except can.CanError:
                    # TODO: specifically implemented for a transmit queue
                    #       full situation to avoid infinite dialogs
                    self.set_bus()

                # TODO: get a real value for sent, but for now python-can
                #       doesn't provide that info.  also, it would be async...
                sent = True

                time.sleep(0.0005)

                if sent:
                    self.tx_notifier.message_received(message=msg)

                    if on_success is not None:
                        on_success()
            else:
                # TODO: I would use message=message (or msg=msg) but:
                #       https://bitbucket.org/hardbyte/python-can/issues/52/inconsistent-send-signatures
                sent = self.bus._send(msg, on_success=on_success, passive=passive)

            if self.auto_disconnect:
                self.verify_bus_ok()

            # TODO: since send() doesn't always report failures this won't either
            #       fix that
            return sent

        return False

    def verify_bus_ok(self):
        if self.bus is None:
            # No bus, nothing to go wrong with it... ?
            ok = True
        else:
            if hasattr(self.bus, "StatusOk"):
                ok = self.bus.StatusOk()

                if not ok:
                    self.set_bus()
            else:
                ok = True
                # TODO: support socketcan
                if hasattr(self.bus, "verify_bus_ok"):
                    ok = self.bus.verify_bus_ok()

        return ok

    def shutdown(self):
        pass

    def flash(self):
        if self.bus is not None:
            result = self.bus.flash(True)
            self.flash_timer.start()
            return result

    def stop_flashing(self):
        if self.bus is not None:
            return self.bus.flash(False)

    def terminate(self):
        self.set_bus()
        logging.debug("{} terminated".format(object.__repr__(self)))

    def set_bus(self, bus=None):
        was_online = self.bus is not None

        if was_online:
            if isinstance(self.bus, can.BusABC):
                self.real_notifier.stop()
                time.sleep(1.1 * self.timeout)
            else:
                self.bus.notifier.remove(self.notifier)
                self.bus.tx_notifier.remove(self.tx_notifier)
            self.bus.shutdown()

        if isinstance(bus, can.BusABC):
            self.bus_settings = BusSettings.from_bus(real_bus=bus)
        elif bus is not None:
            # is being set to a proxy
            self.bus_settings = None

        self.bus = bus

        if self.bus is not None:
            self.set_filters(self.filters)
            if isinstance(self.bus, can.BusABC):
                self.real_notifier = can.Notifier(
                    bus=self.bus, listeners=[self.notifier], timeout=self.timeout
                )
            else:
                self.bus.notifier.add(self.notifier)
                self.bus.tx_notifier.add(self.tx_notifier)
                self.real_notifier = None
        else:
            self.real_notifier = None

        self.reset()

        if was_online and self.bus is None:
            self.went_offline.emit()

        if isinstance(self.bus, can.BusABC):
            self.notifier.move_to_thread(None)
            self.tx_notifier.move_to_thread(None)
        else:
            app = QApplication.instance()
            if app is not None:
                self.notifier.move_to_thread(app.thread())
                self.tx_notifier.move_to_thread(app.thread())

    def reset(self):
        if self.bus is not None:
            if isinstance(self.bus, can.interfaces.pcan.PcanBus):
                self.bus.reset()
                # TODO: do this a better way
                # Give PCAN a chance to actually reset and avoid immediate
                # send failures
                time.sleep(0.500)
            else:
                # TODO: support socketcan
                if hasattr(self.bus, "reset"):
                    self.bus.reset()

    def inner_proxy(self):
        maybe = self
        while True:
            if maybe.bus is None or isinstance(maybe.bus, can.BusABC):
                return maybe

            maybe = maybe.bus

    def reconnect(self):
        inner = self.inner_proxy()

        inner.set_bus()
        inner.set_bus(bus=self.bus_settings.create_bus())

    def set_filters(self, filters):
        self.filters = filters
        real_bus = self.bus
        if real_bus is not None:
            if hasattr(real_bus, "setFilters"):
                real_bus.setFilters(can_filters=filters)


class NotifierProxy(QtCanListener):
    def __init__(self, bus, listeners=[], filtered_ids=None, parent=None):
        super().__init__(receiver=self.message_received, parent=parent)

        # TODO: consider a WeakSet, though this may presently
        #       be keeping objects alive
        self.listeners = set(listeners)
        if filtered_ids is None:
            self.filtered_ids = None
        else:
            self.filtered_ids = set(filtered_ids)

    def message_received(self, message):
        if self.filtered_ids is None or message.arbitration_id in self.filtered_ids:
            for listener in tuple(self.listeners):
                listener.message_received_signal.emit(message)

    def add(self, listener):
        self.listeners.add(listener)

    def discard(self, listener):
        self.listeners.discard(listener)

    def remove(self, listener):
        self.listeners.remove(listener)


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

from epyqlib.iopoint import IoPoint
from PyQt5.QtCore import pyqtProperty
from PyQt5 import QtWidgets

import epyqlib.iogroup_ui
import epyqlib.utils.qt


# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class IoGroup(QtWidgets.QWidget):
    def __init__(self, parent=None, in_designer=False):
        super().__init__(parent=parent)

        self.in_designer = in_designer

        self._status_frame = ""
        self._status_signal_format = ""
        self._set_frame = ""
        self._set_signal_format = ""
        self._override_frame = ""
        self._override_signal_format = ""

        self._quantity = 0
        self._tx = True

        self.points = []

        self.ui = epyqlib.iogroup_ui.Ui_Form()
        self.ui.setupUi(self)

    @pyqtProperty(bool)
    def tx(self):
        return self._tx

    @tx.setter
    def tx(self, tx):
        self._tx = bool(tx)

        self.update_configuration()

    @pyqtProperty(str)
    def box_title(self):
        return self.ui.box.title()

    @box_title.setter
    def box_title(self, title):
        self.ui.box.setTitle(title)

    @pyqtProperty(int)
    def quantity(self):
        return self._quantity

    @quantity.setter
    def quantity(self, quantity):
        self._quantity = quantity

        self.update_configuration()

    @pyqtProperty(str)
    def status_frame(self):
        return self._status_frame

    @status_frame.setter
    def status_frame(self, frame):
        self._status_frame = frame

        self.update_configuration()

    @pyqtProperty(str)
    def status_signal_format(self):
        return self._status_signal_format

    @status_signal_format.setter
    def status_signal_format(self, signal_format):
        self._status_signal_format = signal_format

        self.update_configuration()

    @pyqtProperty(str)
    def set_frame(self):
        return self._set_frame

    @set_frame.setter
    def set_frame(self, frame):
        self._set_frame = frame

        self.update_configuration()

    @pyqtProperty(str)
    def set_signal_format(self):
        return self._set_signal_format

    @set_signal_format.setter
    def set_signal_format(self, signal_format):
        self._set_signal_format = signal_format

        self.update_configuration()

    @pyqtProperty(str)
    def override_frame(self):
        return self._override_frame

    @override_frame.setter
    def override_frame(self, frame):
        self._override_frame = frame

        self.update_configuration()

    @pyqtProperty(str)
    def override_signal_format(self):
        return self._override_signal_format

    @override_signal_format.setter
    def override_signal_format(self, signal_format):
        self._override_signal_format = signal_format

        self.update_configuration()

    def update_configuration(self):
        while self.points:
            point = self.points.pop()
            self.ui.layout.removeWidget(point)
            point.setParent(None)

        for i in range(self.quantity):
            point = IoPoint(parent=self, in_designer=self.in_designer)
            point.index = i + 1
            point.tx = self._tx
            point.status_frame = self.status_frame
            point.status_signal = self.status_signal_format.format(point.index)
            point.set_frame = self.set_frame
            point.set_signal = self.set_signal_format.format(point.index)
            point.override_frame = self.override_frame
            point.override_signal = self.override_signal_format.format(point.index)
            self.ui.layout.addWidget(point)
            self.points.append(point)


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

from PyQt5.QtCore import pyqtProperty

# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class OverrideRange:
    def __init__(self, *args, **kwargs):
        self._override_range = False
        self._min = 0
        self._max = 1

        super().__init__(*args, **kwargs)

    @pyqtProperty(bool)
    def override_range(self):
        return self._override_range

    @override_range.setter
    def override_range(self, override):
        self._override_range = bool(override)

    @pyqtProperty(float)
    def minimum(self):
        return self._min

    @minimum.setter
    def minimum(self, min):
        self._min = min

    @pyqtProperty(float)
    def maximum(self):
        return self._max

    @maximum.setter
    def maximum(self, max):
        self._max = max


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import collections
import enum
import inspect
import itertools
import logging
import math
import os
import pathlib
import shutil
import stat
import traceback
import textwrap
import time
import zipfile
from typing import Optional

import attr

__copyright__ = "Copyright 2017, EPC Power Corp."
__license__ = "GPLv2+"

logger = logging.getLogger()


class ExpectedException(Exception):
    def expected_message(self):
        return "This is expected: {}".format(str(self))


class UnmatchedSerialNumberError(Exception):
    pass


class Container:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


def multiparagraph_wrap(s, *args, **kwargs):
    paragraphs = s.split("\n\n")
    return "\n\n".join("\n".join(textwrap.wrap(p, *args, **kwargs)) for p in paragraphs)


@attr.s
class AverageValueRate:
    _seconds = attr.ib(converter=float)
    _deque = attr.ib(default=attr.Factory(collections.deque))

    @attr.s
    class Event:
        time = attr.ib()
        value = attr.ib()
        delta = attr.ib()

    def add(self, value):
        now = time.monotonic()

        if len(self._deque) > 0:
            delta = now - self._deque[-1].time

            cutoff_time = now - self._seconds

            while self._deque[0].time < cutoff_time:
                self._deque.popleft()
        else:
            delta = 0

        event = self.Event(time=now, value=value, delta=delta)
        self._deque.append(event)

    def rate(self):
        if len(self._deque) > 0:
            dv = self._deque[-1].value - self._deque[0].value
            dt = self._deque[-1].time - self._deque[0].time
        else:
            dv = -1
            dt = 0

        if dv <= 0:
            return 0
        elif dt == 0:
            return math.inf

        return dv / dt

    def remaining_time(self, final_value):
        rate = self.rate()
        if rate <= 0:
            return math.inf
        else:
            return (final_value - self._deque[-1].value) / rate


def write_device_to_zip(
    zip_path, epc_dir, referenced_files, code=None, sha=None, checkout_dir=None
):
    # TODO: stdlib zipfile can't create an encrypted .zip
    #       make a good solution that will...
    with zipfile.ZipFile(file=zip_path, mode="w") as zip:
        for device_path in referenced_files:
            filename = os.path.join(epc_dir, device_path)
            zip.write(
                filename=filename, arcname=os.path.relpath(filename, start=epc_dir)
            )

        if sha is not None:
            sha_file_name = "sha"
            sha_file_path = os.path.join(checkout_dir, sha_file_name)
            with open(sha_file_path, "w") as sha_file:
                sha_file.write(sha + "\n")
            zip.write(filename=sha_file_path, arcname=sha_file_name)


# https://docs.python.org/3/library/itertools.html
def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


# https://docs.python.org/3/library/itertools.html
def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue=fillvalue)


def chunker(iterable, n, factory=tuple):
    """Collect data into fixed-length chunks or blocks except last due to
    lack of fill.
    """
    # chunker('ABCDEFG', 3) --> ABC DEF G"

    i = iter(iterable)
    r = range(n)

    while True:
        # t = factory(next(i) for _ in r)
        t = factory(v for _, v in zip(r, i))

        if len(t) > 0:
            yield t
        else:
            break


def generate_ranges(ids):
    try:
        start = ids[0]
    except IndexError:
        return

    for previous, next in pairwise(itertools.chain(ids, (None,))):
        if previous + 1 != next:
            yield (start, previous)

            start = next


@attr.s
class append_to_method:
    callable = attr.ib()
    name = attr.ib(default=None)

    def __attrs_post_init__(self):
        if self.name is None:
            self.name = self.callable.__name__

    def __call__(self, cls):
        old = getattr(cls, self.name, None)
        to_call = self.callable

        def f(s, *args, **kwargs):
            if old is not None:
                old(s, *args, **kwargs)

            to_call(s)

        setattr(cls, self.name, f)
        print("assigned {} to {}.{}".format(f, cls.__name__, self.name))

        return cls


# Based on https://docs.python.org/3.5/library/enum.html#autonumber
class AutoNumberIntEnum(enum.IntEnum):
    def __new__(cls, value=None):
        if value is None:
            if len(cls.__members__) == 0:
                base = 0
            else:
                base = tuple(cls.__members__.values())[-1]
            value = base + 1

        obj = int.__new__(cls, value)
        obj._value_ = value
        return obj


def enumerated_attrs(enumeration, **kwargs):
    def inner(cls):
        for enumerator in enumeration:
            setattr(cls, enumerator.name, attr.ib(**kwargs))

        cls.enumeration = enumeration

        return cls

    return inner


class TextTable:
    def __init__(self, format_string=None):
        self.format = format_string
        self.rows = []
        self.widths = ()

    def append(self, *row):
        row = tuple(str(c) for c in row)
        self.rows.append(row)
        widths = tuple(len(c) for c in row)
        self.widths = tuple(
            max(a, b)
            for a, b in itertools.zip_longest(self.widths, widths, fillvalue=0)
        )

    def extend(self, rows):
        for row in rows:
            self.append(*row)

    def __str__(self):
        if len(self.rows) == 0:
            return ""

        f = self.format
        if f is None:
            f = " ".join(("{{:{}}}",) * len(self.widths))

        f = f.format(*(max(w, 1) for w in self.widths))

        return "\n".join(
            (
                f.format(*(row + (("",) * (len(self.widths) - len(row)))))
                for row in self.rows
            )
        )


@attr.s
class Collector(list):
    cls = attr.ib()
    payload_attribute = attr.ib()

    def append(self, *args, **kwargs):
        s = super()

        def f(payload):
            payload_dict = {self.payload_attribute: payload}
            s.append(
                self.cls(
                    *args,
                    **kwargs,
                    **{self.payload_attribute: payload},
                )
            )

            return payload

        return f


def exception_logger(excType, excValue, tracebackobj):
    logger.error(
        "Uncaught exception hooked:\n{}".format(
            "".join(traceback.format_exception(excType, excValue, tracebackobj))
        )
    )


@attr.s
class TypeMap:
    types = attr.ib(default=attr.Factory(dict), init=False)

    def __call__(self, wrapped):
        def inner(cls):
            self.types[wrapped] = cls

            return cls

        return inner

    def __getitem__(self, item):
        return self.types[item]

    def wrap(self, wrapped, *args, **kwargs):
        return self.types[type(wrapped)](wrapped=wrapped, *args, **kwargs)


def spaced_to_camel(name, upper):
    segments = name.split(" ")

    if not upper:
        segments[0] = segments[0].lower()

    segments = itertools.chain(
        segments[:1],
        *(
            "".join(
                itertools.chain(
                    c[:1].upper(),
                    c[1:],
                )
            )
            for c in segments[1:]
        ),
    )

    return "".join(segments)


def spaced_to_lower_camel(name):
    return spaced_to_camel(name=name, upper=False)


def spaced_to_upper_camel(name):
    return spaced_to_camel(name=name, upper=True)


def underscored_to_upper_camel(name):
    return "".join(w.title() for w in name.split("_"))


def underscored_camel_to_upper_camel(name):
    return "".join(w[:1].upper() + w[1:] for w in name.split("_"))


def cameled_to_spaced(name):
    word_indexes = sorted(
        (0, len(name))
        + tuple(
            i
            for i, (c, next_c) in enumerate(pairwise(name))
            if c.isupper() and next_c.islower()
        )
        + tuple(
            i + 1
            for i, (prev_c, c) in enumerate(pairwise(name))
            if not prev_c.isupper() and c.isupper()
        )
    )

    words = (name[start:end] for start, end in pairwise(word_indexes) if end > start)

    return " ".join(word.strip() for word in words).strip()


def underscored_camel_to_title_spaced(name):
    result = underscored_camel_to_upper_camel(name)
    result = cameled_to_spaced(result)

    return result


def identifier_path(it):
    return "__" + "_".join(it.__module__.split(".") + [it.__qualname__])


# TODO: CAMPid 0238493420143087667542054268097120437916848
# http://stackoverflow.com/a/21263493/228539
def del_rw(action, name, exc):
    os.chmod(name, stat.S_IWRITE)
    if os.path.isdir(name):
        os.rmdir(name)
    else:
        os.remove(name)


def rmtree(path, ignore_errors=False):
    try:
        shutil.rmtree(path, onerror=del_rw)
    except FileNotFoundError:
        if not ignore_errors:
            raise


def contiguous_groups(s):
    groups = []
    group = []

    for previous, current in pairwise(itertools.chain(s, (math.inf,))):
        group.append(previous)
        if current > previous + 1:
            groups.append(group)
            group = []

    return groups


def path_and_line(o):
    return "{}:{}".format(
        inspect.getsourcefile(o),
        inspect.getsourcelines(o)[1],
    )


def intersect_dicts(a, b):
    common_keys = set(a).intersection(b)
    return {k: a[k] for k in common_keys}


def safe_get(dct: dict, keys) -> Optional[dict]:
    if type(keys) is not list and type(keys) is not tuple:
        keys = [keys]
    for key in keys:
        try:
            dct = dct[key]
        except (KeyError, IndexError):
            return None

    return dct


# TODO: CAMPid 073407143081341008467657184603164130
def format_nested_lists(it, indent="    ", indentation=""):
    result = []

    for item in it:
        if isinstance(item, list):
            result.extend(
                format_nested_lists(
                    it=item, indent=indent, indentation=indentation + indent
                ),
            )
        elif item.strip() == "":
            result.append("")
        else:
            result.append(indentation + item)

    if indentation == "":
        result.append("")
        return "\n".join(result)
    else:
        return result


def ordered_unique(iterable):
    seen = set()

    for item in iterable:
        if item not in seen:
            seen.add(item)
            yield item


@attr.s
class Orderer:
    ordered = attr.ib()
    indexes = attr.ib()
    key = attr.ib()

    @classmethod
    def build(cls, ordered, key=lambda item: item):
        indexes = {item: index for index, item in enumerate(ordered)}

        return cls(
            ordered=ordered,
            indexes=indexes,
            key=key,
        )

    def __call__(self, value):
        return self.indexes.get(self.key(value), math.inf)



import base64
import click
import datetime
import json
import os
import requests
from selenium import webdriver
from selenium.webdriver.common.keys import Keys
from urllib.parse import urljoin

# The output:
# file_dump: Dump of all of the Phabricator files in the system
# maniphest_dump: Text dump of the Phabricator tasks/tickets in the system
# pdf_dump: PDF representations of the Phabricator tasks
# file_search.txt: Searchable text that provides link between
#   Phabricator maniphest tickets/tasks and the files in file_dump


def extract_files_search(phabricator_url, api_token):
    # Retrieve the files data from Phabricator.
    file_search_request_data = {
        "api.token": api_token,
        "after": None,
    }

    file_search_result_data = []
    while True:
        file_search_url = urljoin(phabricator_url, "/api/file.search")
        response = requests.post(
            file_search_url, data=file_search_request_data, verify=False
        )
        json_response = response.json()
        result_data = json_response["result"]["data"]
        file_search_result_data.extend(result_data)
        result_cursor = json_response["result"]["cursor"]
        next_after = result_cursor["after"]
        next_before = result_cursor["before"]
        if next_after is None:
            # This is a special case that was necessary in order to skip over restricted files.
            # See the files in the comments at the bottom of this file.
            if next_before == "23067":
                next_after = 22840
            else:
                break
        file_search_request_data["after"] = next_after

    return file_search_result_data


def generate_file_search_text_file(target, file_search_result_data):
    # Generate the file_search.txt file, which provides a link between the Maniphest tickets and the files.
    file_name_out = os.path.join(target, "file_search.txt")
    with open(file_name_out, "w", encoding="utf-8") as file_to_save:
        for result in file_search_result_data:
            file_to_save.write(f"id: {result['id']}\n")
            file_to_save.write(f"phid: {result['phid']}\n")
            file_to_save.write(f"uri: {result['fields']['uri']}\n")
            file_to_save.write(f"name: {result['fields']['name']}\n")
            file_to_save.write(f"type: {result['type']}\n\n")


def download_files(phabricator_url, api_token, target, file_search_result_data):
    # File download from Phabricator given file search results.
    file_download_request_data = {
        "api.token": api_token,
        "phid": None,
    }

    file_output_dir = os.path.join(target, "file_dump")
    for result_data in file_search_result_data:
        id = result_data["id"]
        file_name = result_data["fields"]["name"]
        if file_name is None or file_name == "":
            file_name = "BLANK_FILE_NAME"
        file_name_out = os.path.join(file_output_dir, str(id), file_name)
        os.makedirs(os.path.dirname(file_name_out), exist_ok=True)
        file_download_request_data["phid"] = result_data["phid"]
        print(f"File download: {id} {file_name}")
        file_download_url = urljoin(phabricator_url, "/api/file.download")
        response = requests.post(
            file_download_url, data=file_download_request_data, verify=False
        )
        file_data = response.json()["result"]
        if file_data is not None:
            data_bytes = file_data.encode("utf-8")
        else:
            print("WARNING: DATA IS NONE")
            data_bytes = "".encode("utf-8")
            readme_file_name_out = os.path.join(
                file_output_dir, str(id), "readme_phabricator_extract.txt"
            )
            with open(readme_file_name_out, "w") as readme_file_to_save:
                readme_file_to_save.write(f"id: {id}\n")
                readme_file_to_save.write(f"file name: {file_name}\n")
                readme_file_to_save.write(f"phid: {result_data['phid']}\n")
                readme_file_to_save.write(
                    "NOTE: This file was written because the contents of the file were either empty or unable to be read."
                )
        with open(file_name_out, "wb") as file_to_save:
            decoded_data = base64.decodebytes(data_bytes)
            file_to_save.write(decoded_data)


def extract_file_data(phabricator_url, api_token, target):
    # Extract files from Phabricator and store in target directory.
    file_search_result_data = extract_files_search(phabricator_url, api_token)
    generate_file_search_text_file(target, file_search_result_data)
    download_files(phabricator_url, api_token, target, file_search_result_data)


def get_users(phabricator_url, api_token):
    # Retrieve user information from Phabricator.
    user_search_request_data = {
        "api.token": api_token,
        "after": None,
    }

    total_result_data = []
    user_search_url = urljoin(phabricator_url, "/api/user.search")

    while True:
        response = requests.post(
            user_search_url, data=user_search_request_data, verify=False
        )
        json_response = response.json()
        result_data = json_response["result"]["data"]
        total_result_data.extend(result_data)
        result_cursor = json_response["result"]["cursor"]
        next_after = result_cursor["after"]
        if next_after is None:
            break
        user_search_request_data["after"] = next_after

    user_dict = {}
    for result_data in total_result_data:
        user_phid = result_data["phid"]
        user_name = result_data["fields"]["realName"]
        user_dict[user_phid] = user_name

    return user_dict


def extract_maniphest_search(phabricator_url, api_token):
    # Retrieve Maniphest ticket/task information.
    maniphest_search_request_data = {
        "api.token": api_token,
        "after": None,
    }

    maniphest_search_result_data = []
    maniphest_search_url = urljoin(phabricator_url, "/api/maniphest.search")

    while True:
        response = requests.post(
            maniphest_search_url, data=maniphest_search_request_data, verify=False
        )
        json_response = response.json()
        result_data = json_response["result"]["data"]
        maniphest_search_result_data.extend(result_data)
        result_cursor = json_response["result"]["cursor"]
        next_after = result_cursor["after"]
        if next_after is None:
            break
        maniphest_search_request_data["after"] = next_after

    return maniphest_search_result_data


def write_maniphest_data_to_text_file(target, maniphest_search_result_data, user_dict):
    # Write Maniphest ticket/task information to searchable text file.
    file_output_dir = os.path.join(target, "maniphest_dump")
    os.makedirs(file_output_dir, exist_ok=True)
    maniphest_id_list = []
    for result_data in maniphest_search_result_data:
        maniphest_id = result_data["id"]
        maniphest_id_list.append(maniphest_id)

        maniphest_file_name = os.path.join(file_output_dir, str(maniphest_id) + ".txt")
        with open(maniphest_file_name, "w", encoding="utf-8") as maniphest_fp:
            maniphest_fp.write(f"name: {result_data['fields']['name']}\n")
            maniphest_fp.write(f"status: {result_data['fields']['status']['name']}\n")
            maniphest_fp.write(
                f"priority: {result_data['fields']['priority']['name']}\n"
            )
            maniphest_fp.write(
                f"date created: {datetime.datetime.fromtimestamp(result_data['fields']['dateCreated'])}\n"
            )
            maniphest_fp.write(
                f"date modified: {datetime.datetime.fromtimestamp(result_data['fields']['dateModified'])}\n"
            )
            maniphest_fp.write(
                f"author: {user_dict.get(result_data['fields']['authorPHID'])}\n"
            )
            maniphest_fp.write(
                f"owner: {user_dict.get(result_data['fields']['ownerPHID'])}\n"
            )
            maniphest_fp.write(
                f"description: {result_data['fields']['description']['raw']}\n"
            )
            if len(result_data["attachments"]) > 0:
                maniphest_fp.write(f"attachments: {result_data['attachments']}\n")


def maniphest_to_pdf(
    maniphest_search_result_data,
    target,
    chromedriver,
    phabricator_url,
    username,
    password,
):
    # PDF dump of Maniphest ticket/task web pages.
    file_output_dir = os.path.join(target, "pdf_dump")

    options = webdriver.ChromeOptions()

    settings = {
        "recentDestinations": [
            {
                "id": "Save as PDF",
                "origin": "local",
                "account": "",
            }
        ],
        "selectedDestinationId": "Save as PDF",
        "version": 2,
    }
    # TODO: Never could get command to save in the specified directory. Always saves to Downloads dir.
    prefs = {
        "profile.default_content_settings.popups": 0,
        "download.default_directory": file_output_dir + os.path.sep,
        "download.prompt_for_download": False,
        "download.directory_upgrade": True,
        "printing.print_preview_sticky_settings.appState": json.dumps(settings),
    }
    options.add_experimental_option("prefs", prefs)
    options.add_argument("--kiosk-printing")

    options.add_argument("--ignore-ssl-errors=yes")
    options.add_argument("--ignore-certificate-errors")
    # TODO: automatically load/install chromedriver
    driver = webdriver.Chrome(executable_path=chromedriver, options=options)
    driver.get(phabricator_url)

    elem = driver.find_element_by_name("username")
    elem.clear()
    elem.send_keys(username)

    elem = driver.find_element_by_name("password")
    elem.clear()
    elem.send_keys(password)

    elem.send_keys(Keys.RETURN)

    os.makedirs(file_output_dir, exist_ok=True)
    for result_data in maniphest_search_result_data:
        maniphest_id = result_data["id"]
        maniphest_url = urljoin(phabricator_url, "/T" + str(maniphest_id))
        driver.get(maniphest_url)
        # TODO: Never could get command to save in the specified directory. Always saves to Downloads dir.
        driver.execute_script(f"window.print();")

    driver.close()


def extract_tickets(
    chromedriver, phabricator_url, api_token, target, username, password
):
    user_dict = get_users(phabricator_url, api_token)
    maniphest_search_result_data = extract_maniphest_search(phabricator_url, api_token)
    write_maniphest_data_to_text_file(target, maniphest_search_result_data, user_dict)
    maniphest_to_pdf(
        maniphest_search_result_data,
        target,
        chromedriver,
        phabricator_url,
        username,
        password,
    )


def create_command():
    @click.command()
    @click.option(
        "--target",
        help="path to the output directory",
        type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True),
        required=True,
    )
    @click.option(
        "--chromedriver",
        help="path to chromedriver",
        type=click.Path(exists=True, file_okay=True, resolve_path=True),
        required=True,
    )
    @click.option(
        "--phabricator-url",
        help="The base URL for the Phabricator server.",
        type=str,
        required=True,
    )
    @click.option(
        "--api-token",
        help="""
            Your Phabricator API token.
            These are managed under your user settings then 'Conduit API Tokens'.
        """,
        type=str,
        required=True,
    )
    @click.option(
        "--username",
        help="Your Phabricator user name.",
        type=str,
        required=True,
    )
    @click.password_option()
    def cli(target, chromedriver, phabricator_url, api_token, username, password):
        extract_file_data(phabricator_url, api_token, target)
        extract_tickets(
            chromedriver, phabricator_url, api_token, target, username, password
        )

    return cli



from os import path

from epyqlib.tabs.files.files_manager import FilesManager

# noinspection PyUnresolvedReferences
from epyqlib.tests.utils.test_fixtures import temp_dir


def test_hashing(temp_dir):
    empty_file_hash = "d41d8cd98f00b204e9800998ecf8427e"

    manager = FilesManager(temp_dir)
    with open(path.join(manager._cache_dir, "test"), "w") as test:
        test.write("test")

    open(path.join(manager._cache_dir, empty_file_hash), "w").close()

    manager.verify_cache()

    assert empty_file_hash in manager.hashes()
    assert "test" not in manager.hashes()



import inspect
import json
from urllib.parse import urlparse

import paho.mqtt.client as mqtt
from twisted.internet.defer import ensureDeferred
from twisted.internet.task import LoopingCall
from typing import Callable, Coroutine, List, Dict

SocketCloseHandler = Callable[[], Coroutine]
OnMessageHandler = Callable[[str, Dict], Coroutine]


class WebSocketHandler:
    def __init__(self):
        self._tag = "[Graphql Websocket]"
        self._on_message_handler: OnMessageHandler = None
        self._on_close_handler: SocketCloseHandler = None
        self.loopingCall: LoopingCall = None
        self.clients: List[mqtt.Client] = []
        self.disconnecting = False

        self.resubscribe = False

    def connect(
        self,
        response: Dict,
        on_message: Callable[[str, Dict], Coroutine],
        on_close: SocketCloseHandler = None,
    ) -> None:
        """
        Makes WebSocket connection and starts looping
        :param response JSON body response from a subscription call to AppSync
        :param on_message: Coroutine that is called on every message. The first param is the action (created/updated/deleted)
        and the second is the file json object as a dict.
        :param on_close: Coroutine that is called when any subscribed socket is closed.
        """
        self._on_message_handler = on_message
        self._on_close_handler = on_close

        new_subscriptions = response["extensions"]["subscription"]["newSubscriptions"]
        mqtt_connections = response["extensions"]["subscription"]["mqttConnections"]

        new_connections = {}
        for [action, details] in new_subscriptions.items():
            mqtt_connection = next(
                c for c in mqtt_connections if details["topic"] in c["topics"]
            )
            topic = details["topic"]

            if mqtt_connection["url"] not in new_connections:
                new_connections[mqtt_connection["url"]] = {
                    "connection": mqtt_connection,
                    "topics": set(),
                }

            new_connections[mqtt_connection["url"]]["topics"].add(topic)

        self._mqtt_connections = new_connections
        self._do_connect()

        self.resubscribe = True

        self.loopingCall: LoopingCall = LoopingCall(self.loop)
        self.loopingCall.start(1)

    def _do_connect(self):
        for connection in self._mqtt_connections.values():
            client_id = connection["connection"]["client"]
            url = connection["connection"]["url"]

            urlparts = urlparse(url)

            headers = {"Host": "{0:s}".format(urlparts.netloc)}

            client = mqtt.Client(client_id=client_id, transport="websockets")
            client.on_connect = self._on_connect
            client.on_message = self._on_message
            client.on_log = self._on_log
            client.on_socket_close = self._on_socket_close

            client.ws_set_options(
                path="{}?{}".format(urlparts.path, urlparts.query), headers=headers
            )
            client.tls_set()

            client.user_data_set({"topics": connection["topics"]})

            client.connect(urlparts.netloc, port=443)
            self.clients.append(client)

    def loop(self):
        # if not self.is_subscribed():
        #     # If we shouldn't resub, stop looping
        #     if not self.resubscribe:
        #         self.loopingCall.stop()
        #         return
        #
        #     # Otherwise, re-sub
        #     self._do_connect()

        if len(self.clients) == 0:
            self.loopingCall.stop()

        for client in self.clients:
            client.loop_read()
            client.loop_write()
            client.loop_misc()

    def disconnect(self):
        self.resubscribe = False
        for client in self.clients:
            client.disconnect()

    def is_subscribed(self):
        return len(self.clients) > 0

    def _on_connect(self, client: mqtt.Client, userdata: Dict, flags, rc):
        topics: set[str] = userdata["topics"]
        sub_list = list(map(lambda topic: (topic, 1), topics))
        client.subscribe(sub_list)

    def _on_log(self, client, userdata, level, buf):
        print(f"{self._tag}  Log {level} {buf}")

    def _on_message(self, client, userdata, msg: mqtt.MQTTMessage):
        try:
            payload = msg.payload.decode("ascii")
            payload_json = json.loads(payload)
            print(f"{self._tag} Message received: {payload}")
        except Exception as e:
            print(
                f"{self._tag} Error converting payload to JSON: "
                + msg.payload.decode("ascii")
            )
            print(e)
            return

        # We *should* only get one payload, but just in case...
        try:
            for action, payload in payload_json["data"].items():
                result = self._on_message_handler(action, payload)
                if inspect.iscoroutine(result):
                    ensureDeferred(result)
        except Exception as e:
            print(
                f"{self._tag} Error iterating over payload: " + json.dumps(payload_json)
            )
            print(e)

    def _on_socket_close(self, client: mqtt.Client, userdata: Dict, socket):
        if self.resubscribe:
            status = client.reconnect()
            if status == 0:
                return
            else:
                print(
                    f"{self._tag} Error reconnecting to websocket: {self.error_lookup[status]}. Aborting reconnect."
                )

        self.clients.remove(client)
        # print(f"Socket closed. Connection to topics ${userdata.get('topics')} closed.")
        # self.clients.remove(client)
        #
        # # Make sure this only fires once no matter how many clients are open
        # if self.disconnecting is False:
        #     self.disconnecting = True
        #     self.disconnect()
        #     self.disconnecting = False
        #
        #
        #     # Make sure the on_close handler only gets run once
        #     if self._on_close_handler:
        #         result = self._on_close_handler()
        #         if inspect.iscoroutine(result):
        #             ensureDeferred(result)

    # From https://pypi.org/project/paho-mqtt/#on-connect
    error_lookup = {
        mqtt.CONNACK_ACCEPTED: "Connection successful",
        mqtt.CONNACK_REFUSED_PROTOCOL_VERSION: "Connection refused - incorrect protocol version",
        mqtt.CONNACK_REFUSED_IDENTIFIER_REJECTED: "Connection refused - invalid client",
        mqtt.CONNACK_REFUSED_SERVER_UNAVAILABLE: "Connection refused - server unavailable",
        mqtt.CONNACK_REFUSED_BAD_USERNAME_PASSWORD: "Connection refused - bad username or password",
        mqtt.CONNACK_REFUSED_NOT_AUTHORIZED: "Connection refused - not authorised",
    }


# Example of the format of `response`:
# {
#   "extensions": {
#     "subscription": {
#       "mqttConnections": [
#         {
#           "url": "wss://a1yyia7sgxh08y-ats.iot.us-west-2.amazonaws.com/mqtt?X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIA5..."
#           "topics": [
#             "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/associationCreated/ad2fc514666b327b3e79f4255329988b30bfec23428bd62f7db900ffce4744a7",
#             "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/activityCreated/38c53ba841180e186aeaae8565e9807ddabb7becaf849cf336e24d1f23e2a12b",
#             "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/fileCreated/ad2fc514666b327b3e79f4255329988b30bfec23428bd62f7db900ffce4744a7"
#           ],
#           "client": "xtsaum2eyfhaxd2scqsjnajvmq"
#         },
#         {
#           "url": "wss://a1yyia7sgxh08y-ats.iot.us-west-2.amazonaws.com/mqtt?X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIA5..."
#           "topics": [
#             "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/fileUpdated/65a5e5608f9c9ce6e418848ef29015d110c059d1d02fb103288f5a7bab394e72",
#             "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/associationDeleted/65a5e5608f9c9ce6e418848ef29015d110c059d1d02fb103288f5a7bab394e72",
#             "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/fileUpdated/ad2fc514666b327b3e79f4255329988b30bfec23428bd62f7db900ffce4744a7"
#           ],
#           "client": "bqaskf7ddrfrnctwj4mf7fkwpi"
#         },
#         {
#           "url": "wss://a1yyia7sgxh08y-ats.iot.us-west-2.amazonaws.com/mqtt?X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIA5..."
#           "topics": [
#             "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/fileDeleted/129baf6c447f4da79303537da82ec8e71266c21536567ded0f5d997f75e889e3",
#             "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/fileDeleted/65a5e5608f9c9ce6e418848ef29015d110c059d1d02fb103288f5a7bab394e72",
#             "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/associationDeleted/ad2fc514666b327b3e79f4255329988b30bfec23428bd62f7db900ffce4744a7"
#           ],
#           "client": "y6myssziyffsvg7e4yfjjngaqy"
#         },
#         {
#           "url": "wss://a1yyia7sgxh08y-ats.iot.us-west-2.amazonaws.com/mqtt?X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIA5..."
#           "topics": [
#             "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/fileUpdated/",
#             "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/associationCreated/65a5e5608f9c9ce6e418848ef29015d110c059d1d02fb103288f5a7bab394e72",
#             "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/fileDeleted/"
#           ],
#           "client": "cbh5lbimwje2nawp42uwmp7o2e"
#         },
#         {
#           "url": "wss://a1yyia7sgxh08y-ats.iot.us-west-2.amazonaws.com/mqtt?X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIA5..."
#           "topics": [
#             "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/fileCreated/129baf6c447f4da79303537da82ec8e71266c21536567ded0f5d997f75e889e3",
#             "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/fileCreated/",
#             "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/fileDeleted/ad2fc514666b327b3e79f4255329988b30bfec23428bd62f7db900ffce4744a7"
#           ],
#           "client": "ollqguz5vvfn3jmyj5tsulz7su"
#         },
#         {
#           "url": "wss://a1yyia7sgxh08y-ats.iot.us-west-2.amazonaws.com/mqtt?X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Credential=ASIA5..."
#           "topics": [
#             "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/fileCreated/65a5e5608f9c9ce6e418848ef29015d110c059d1d02fb103288f5a7bab394e72",
#             "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/fileUpdated/129baf6c447f4da79303537da82ec8e71266c21536567ded0f5d997f75e889e3"
#           ],
#           "client": "tt7sop2eurffzhej45d4gu6ztm"
#         }
#       ],
#       "newSubscriptions": {
#         "activityCreated": {
#           "topic": "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/activityCreated/38c53ba841180e186aeaae8565e9807ddabb7becaf849cf336e24d1f23e2a12b",
#           "expireTime": 1553801051000
#         },
#         "fileUpdated": {
#           "topic": "674475255666/hmdhwjgyuja6nfrcb65tzc42vu/fileUpdated/",
#           "expireTime": 1553801051000
#         }
#       }
#     }
#   },
#   "data": {
#     "activityCreated": null,
#     "fileUpdated": null
#   }
# }



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import epyqlib.widgets.abstracttxwidget
import epyqlib.widgets.button_ui
from PyQt5.QtCore import pyqtProperty


# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class Button(epyqlib.widgets.abstracttxwidget.AbstractTxWidget):
    def __init__(self, parent=None, in_designer=False):
        self._frame = None
        self._signal = None
        self._on_value = 1
        self._off_value = 0

        super().__init__(
            ui_class=epyqlib.widgets.button_ui.Ui_Form,
            parent=parent,
            in_designer=in_designer,
        )

        # TODO: CAMPid 398956661298765098124690765
        self.ui.value.pressed.connect(self.pressed)
        self.ui.value.released.connect(self.released)

    @pyqtProperty(int)
    def on_value(self):
        return self._on_value

    @on_value.setter
    def on_value(self, new_on_value):
        self._on_value = int(new_on_value)

    @pyqtProperty(int)
    def off_value(self):
        return self._off_value

    @off_value.setter
    def off_value(self, new_off_value):
        self._off_value = int(new_off_value)
        self.set(self.off_value)

    def set_signal(self, signal=None, force_update=False):
        super().set_signal(signal, force_update=force_update)

        if signal is not None:
            self.set(self.off_value)

            def get_text_width(widget, text):
                return widget.fontMetrics().boundingRect(text).width()

            button = self.ui.value
            # TODO: it would be nice to use the 'normal' extra width
            # initial_margin = button.width() - get_text_width(button,
            #                                                  button.text())

            if len(self.signal_object.enumeration):
                widths = []
                for text in [
                    self.calculate_text(v) for v in self.signal_object.enumeration
                ]:
                    widths.append(get_text_width(button, text))

                button.setMinimumWidth(1.3 * max(widths))
        else:
            if self.ui is not None:
                self.ui.value.setText("")

    def set(self, value):
        self.widget_value_changed(value)
        self.set_text(value)

    def calculate_text(self, value):
        if self.label_visible and self.signal_object is not None:
            # TODO: CAMPid 85478672616219005471279
            enum_string = self.signal_object.enumeration[value]
            text = self.signal_object.enumeration_format_re["format"].format(
                s=enum_string, v=value
            )

            return text
        else:
            return self.ui.label.text()

    def set_text(self, value):
        self.ui.value.setText(self.calculate_text(value))

    def pressed(self):
        self.set(self.on_value)

    def released(self):
        self.set(self.off_value)

    def set_value(self, value):
        # TODO  exception?
        pass

    def set_unit_text(self, units):
        pass

    def showEvent(self, event):
        epyqlib.widgets.abstracttxwidget.AbstractTxWidget.showEvent(self, event)
        self.set(self.off_value)


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import epyqlib.widgets.abstractsignalpluginclass
import epyqlib.widgets.horizontalslider

# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class HorizontalSliderPlugin(
    epyqlib.widgets.abstractsignalpluginclass.AbstractSignalPlugin
):
    def __init__(self, parent=None):
        epyqlib.widgets.abstractsignalpluginclass.AbstractSignalPlugin.__init__(
            self, parent=parent
        )

        self._init = epyqlib.widgets.horizontalslider.HorizontalSlider
        self._module_path = "epyqlib.widgets.horizontalslider"
        self._name = "HorizontalSlider"


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import json
import os

from typing import List, Dict, Set

from epyqlib.tabs.files.files_utils import ensure_dir


class AssociationCache:

    _instance: "AssociationCache" = None
    _tag = "[Associations Cache]"

    def __init__(self, files_dir: str):
        self._cache_dir = os.path.join(files_dir, "")
        self._cache_file = os.path.join(self._cache_dir, "associations-cache.json")
        ensure_dir(self._cache_dir)

        self._associations: Dict[str, List] = {}

    @staticmethod
    def get_instance():
        if AssociationCache._instance is None:
            raise Exception("AssociationsCache being used before initialized")
        return AssociationCache._instance

    @staticmethod
    def init(files_dir: str):
        AssociationCache._instance = AssociationCache(files_dir)
        AssociationCache._instance._init()
        return AssociationCache._instance

    def _init(self):
        self._read_file()

    def _read_file(self):
        if os.path.exists(self._cache_file):
            with open(self._cache_file, "r") as cache_file:
                self._associations = json.load(cache_file)

    def _write_file(self):
        with open(self._cache_file, "w") as cache_file:
            json.dump(self._associations, cache_file, indent=2)

    def clear(self):
        self._associations = {}
        if os.path.exists(self._cache_file):
            os.unlink(self._cache_file)

    def get_associations(self, serial_number: str) -> List:
        return self._associations.get(serial_number)

    def put_associations(self, serial_number: str, associations: List):
        self._associations[serial_number] = associations
        self._write_file()

    def get_all_known_file_hashes(self) -> Set[str]:
        hashes = set()

        for serial, association_list in self._associations.items():
            for association in association_list:
                if association["file"] is None:
                    continue

                if association["file"]["type"] == "Log":
                    continue

                hashes.add(association["file"]["hash"])

        return hashes
